# AUTO GENERATED ON 2026-02-08 AT 18:23:02
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import cupy.testing as cpt
import numpy as np
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cudaawkward_ListArray64_combinations_length_64_1():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    stops = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [9]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 1, 5, 9]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_2():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    stops = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [139.0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 84.0, 104.0, 139.0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_3():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    stops = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_4():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    stops = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [4]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 0, 4, 4]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_5():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    stops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [3]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_6():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = False
    starts = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    stops = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_7():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = False
    starts = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    stops = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [49.0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 35.0, 39.0, 49.0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_8():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = False
    starts = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    stops = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_9():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = False
    starts = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    stops = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_10():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = False
    starts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    stops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_11():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    stops = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [9]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 1, 5, 9]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_12():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    stops = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [139.0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 84.0, 104.0, 139.0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_13():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    stops = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_14():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    stops = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [4]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 0, 4, 4]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_15():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    stops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [3]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_16():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    stops = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [9]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 1, 5, 9]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_17():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    stops = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [139.0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 84.0, 104.0, 139.0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_18():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    stops = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_19():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    stops = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [4]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 0, 4, 4]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_20():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    stops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [3]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_21():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    stops = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [9]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 1, 5, 9]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_22():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    stops = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [139.0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 84.0, 104.0, 139.0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_23():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    stops = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [0]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 0, 0, 0]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_24():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    stops = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [4]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 0, 4, 4]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cudaawkward_ListArray64_combinations_length_64_25():
    totallen = cupy.array([123], dtype=cupy.int64)
    tooffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    n = 3
    replacement = True
    starts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    stops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_ListArray_combinations_length', cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(totallen, tooffsets, n, replacement, starts, stops, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totallen = [3]
    cpt.assert_allclose(totallen[:len(pytest_totallen)], cupy.array(pytest_totallen))
    pytest_tooffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function

__all__ = ("array_equal",)


@ak._connect.numpy.implements("array_equal")
@high_level_function()
def array_equal(
    a1,
    a2,
    equal_nan: bool = False,
    dtype_exact: bool = True,
    same_content_types: bool = True,
    check_parameters: bool = True,
    check_regular: bool = True,
    check_named_axis: bool = True,
):
    """
    Args:
        a1: Array-like data (anything #ak.to_layout recognizes).
        a2: Array-like data (anything #ak.to_layout recognizes).
        equal_nan: bool (default=False)
            Whether to count NaN values as equal to each other.
        dtype_exact: bool (default=True) whether the dtypes must be exactly the same, or just the
            same family.
        same_content_types: bool (default=True)
            Whether to require all content classes to match
        check_parameters: bool (default=True) whether to compare parameters.
        check_regular: bool (default=True) whether to consider ragged and regular dimensions as
            unequal.
        check_named_axis: bool (default=True) whether to consider named axes as unequal.

    True if two arrays have the same shape and elements, False otherwise.

    TypeTracer arrays are not supported, as there is very little information to
    be compared.
    """
    # Dispatch
    yield a1, a2

    return ak.operations.ak_almost_equal._impl(
        a1,
        a2,
        rtol=0.0,
        atol=0.0,
        dtype_exact=dtype_exact,
        check_parameters=check_parameters,
        check_regular=check_regular,
        check_named_axis=check_named_axis,
        exact_eq=True,
        same_content_types=same_content_types and check_regular,
        equal_nan=equal_nan,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("from_arrow",)

np = NumpyMetadata.instance()


@high_level_function()
def from_arrow(
    array, *, generate_bitmasks=False, highlevel=True, behavior=None, attrs=None
):
    """
    Args:
        array (`pyarrow.Array`, `pyarrow.ChunkedArray`, `pyarrow.RecordBatch`, or `pyarrow.Table`):
            Apache Arrow array to convert into an  Awkward Array.
        generate_bitmasks (bool): If enabled and Arrow/Parquet does not have Awkward
            metadata, `generate_bitmasks=True` creates empty bitmasks for nullable
            types that don't have bitmasks in the Arrow/Parquet data, so that the
            Form (BitMaskedForm vs UnmaskedForm) is predictable.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Converts an Apache Arrow array into an Awkward Array.

    This function always preserves the values of a dataset; i.e. the Python objects
    returned by #ak.to_list are identical to the Python objects returned by Arrow's
    `to_pylist` method. If #ak.to_arrow was invoked with `extensionarray=True`, this
    function also preserves the data type (high-level #ak.types.Type, though not the
    low-level #ak.forms.Form), even through Parquet, making Parquet a good way to save
    Awkward Arrays for later use.

    Because awkward uses numpy's dtype system, timestamp types do not have timezones.
    If encountering timestamp types with timezones in the input arrow data, they
    will be silently dropped.

    See also #ak.to_arrow, #ak.to_arrow_table, #ak.from_parquet, #ak.from_arrow_schema.
    """
    return _impl(array, generate_bitmasks, highlevel, behavior, attrs)


def _impl(array, generate_bitmasks, highlevel, behavior, attrs):
    import awkward._connect.pyarrow

    pyarrow = awkward._connect.pyarrow.pyarrow

    ctx = HighLevelContext(behavior=behavior, attrs=attrs).finalize()

    out = awkward._connect.pyarrow.handle_arrow(
        array, generate_bitmasks=generate_bitmasks, pass_empty_field=True
    )

    if isinstance(array, (pyarrow.lib.Array, pyarrow.lib.ChunkedArray)):
        (
            awkwardarrow_type,
            _storage_type,
        ) = awkward._connect.pyarrow.to_awkwardarrow_storage_types(array.type)

        if awkwardarrow_type is None:
            if isinstance(out, ak.contents.UnmaskedArray):
                out = awkward._connect.pyarrow.remove_optiontype(out)
        else:
            if awkwardarrow_type.mask_type in (None, "IndexedArray"):
                out = awkward._connect.pyarrow.remove_optiontype(out)

            if awkwardarrow_type.record_is_scalar:
                out = out._getitem_at(0)

    def remove_revertable(layout, **kwargs):
        if hasattr(layout, "__pyarrow_original"):
            del layout.__pyarrow_original

    ak._do.recursively_apply(out, remove_revertable)

    return ctx.wrap(out, highlevel=highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext
from awkward._namedaxis import (
    NAMED_AXIS_KEY,
)
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("without_named_axis",)

np = NumpyMetadata.instance()


@high_level_function()
def without_named_axis(
    array,
    *,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Returns an #ak.Array or #ak.Record (or low-level equivalent, if
    `highlevel=False`) without named axes. This function does not change the
    array in-place.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, highlevel, behavior, attrs)


def _impl(array, highlevel, behavior, attrs):
    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(
            array,
            allow_record=True,
            primitive_policy="pass-through",
            none_policy="pass-through",
            string_policy="pass-through",
        )

    return ctx.without_attr(key=NAMED_AXIS_KEY).wrap(
        layout,
        highlevel=highlevel,
        allow_other=True,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from cuda.compute import ZipIterator, gpu_struct, segmented_reduce

from awkward._nplikes.cupy import Cupy

cupy_nplike = Cupy.instance()
cp = cupy_nplike._module

# Cache for cuda.compute availability
_cuda_compute_available: bool | None = None


def is_available() -> bool:
    global _cuda_compute_available

    if _cuda_compute_available is not None:
        return _cuda_compute_available

    try:
        import cuda.compute  # noqa: F401

        _cuda_compute_available = True
    except ImportError:
        _cuda_compute_available = False

    return _cuda_compute_available


def segmented_sort(
    toptr,
    fromptr,
    length,
    offsets,
    offsetslength,
    parentslength,
    ascending,
    stable,
):
    from cuda.compute import SortOrder, segmented_sort

    cupy_nplike = Cupy.instance()
    cp = cupy_nplike._module

    # Ensure offsets are int64 as expected by segmented_sort
    if offsets.dtype != cp.int64:
        offsets = offsets.astype(cp.int64)

    num_segments = offsetslength - 1
    num_items = int(offsets[-1]) if len(offsets) > 0 else 0

    start_offsets = offsets[:-1]
    end_offsets = offsets[1:]

    order = SortOrder.ASCENDING if ascending else SortOrder.DESCENDING

    segmented_sort(
        fromptr,  # d_in_keys
        toptr,  # d_out_keys
        None,  # d_in_values (not sorting values, just keys)
        None,  # d_out_values
        num_items,  # num_items
        num_segments,  # num_segments
        start_offsets,  # start_offsets_in
        end_offsets,  # end_offsets_in
        order,  # order (ASCENDING or DESCENDING)
        None,  # stream (use default stream)
    )


def parents_to_offsets(parents, parents_length):
    if parents_length == 0:
        return cp.array([0], dtype=parents.dtype)

    # count how many elements belong to each parent
    counts = cp.bincount(parents)

    # add a start offset
    offsets = cp.concatenate([cp.array([0], dtype=counts.dtype), counts.cumsum()])

    return offsets


def local_idx_from_parents(parents, parents_length):
    if parents_length == 0:
        return cp.empty(0, dtype=parents.dtype)

    # mark the beginning of each subarray
    new_group = cp.empty(parents_length, dtype=cp.bool_)
    new_group[0] = True
    new_group[1:] = parents[1:] != parents[:-1]

    # find the start index of each subarray
    group_starts = cp.nonzero(new_group)[0]  # shape = (num_groups,)

    # Assign subarray id (1..#subarray) to each element
    group_id = cp.cumsum(new_group)

    # For each element, the start index of its group
    start_pos = group_starts[group_id - 1]

    # local_index = global_index - start_pos
    return cp.arange(parents_length) - start_pos


# the inputs for this function we get from file ~/awkward/src/awkward/_reducers.py:239, in ArgMax.apply(self, array, parents, starts, shifts, outlength)
def awkward_reduce_argmax(
    result,
    input_data,
    parents_data,
    parents_length,
    outlength,
):
    ak_array = gpu_struct(
        {
            "data": input_data.dtype.type,
            "local_index": cp.int64,
        }
    )

    # compare the values of the arrays
    def max_op(a: ak_array, b: ak_array):
        return a if a.data > b.data else b

    # use a helper function to get the local indices
    # local_indices = local_idx_from_parents(parents_data, parents_length)

    # use global indices instead
    global_indices = cp.arange(0, parents_length + 1, dtype=cp.int64)

    # Combine data and their indices into a single structure
    input_struct = ZipIterator(input_data, global_indices)
    # alternative way
    # input_struct = cp.stack((input_data, global_indices), axis=1).view(ak_array.dtype)

    # Prepare the start and end offsets
    offsets = parents_to_offsets(parents_data, parents_length)
    start_o = offsets[:-1]
    end_o = offsets[1:]

    # Prepare the output array
    _result = result
    _result = cp.concatenate((result, result))
    _result = _result.view(ak_array.dtype)

    # alternative way
    # _result = cp.zeros([outlength], dtype= ak_array.dtype)

    # Initial value for the reduction
    # min value gets transformed to input_data.dtype automatically?
    min = cp.iinfo(cp.int64).min
    h_init = ak_array(min, min)

    # Perform the segmented reduce
    segmented_reduce(input_struct, _result, start_o, end_o, max_op, h_init, outlength)

    # TODO: here converts float to int too, fix this?
    _result = _result.view(cp.int64).reshape(-1, 2)
    _result = _result[:, 1]

    # pass the result outside the function
    result_v = result.view()
    result_v[...] = _result


# this function is called from ~/awkward/src/awkward/_reducers.py:161 (ArgMin.apply())
def awkward_reduce_argmin(
    result,
    input_data,
    parents_data,
    parents_length,
    outlength,
):
    index_dtype = parents_data.dtype
    ak_array = gpu_struct(
        {
            "data": input_data.dtype.type,
            "local_index": index_dtype,
        }
    )

    # compare the values of the arrays
    def min_op(a: ak_array, b: ak_array):
        return a if a.data < b.data else b

    # use a helper function to get the local indices
    # local_indices = local_idx_from_parents(parents_data, parents_length)

    # use global indices instead
    global_indices = cp.arange(0, parents_length + 1, dtype=cp.int64)

    # Combine data and their indices into a single structure
    input_struct = ZipIterator(input_data, global_indices)
    # alternative way
    # input_struct = cp.stack((input_data, global_indices), axis=1).view(ak_array.dtype)

    # Prepare the start and end offsets
    offsets = parents_to_offsets(parents_data, parents_length)
    start_o = offsets[:-1]
    end_o = offsets[1:]

    # Prepare the output array
    _result = result
    _result = cp.concatenate((result, result))
    _result = _result.view(ak_array.dtype)

    # alternative way
    # _result = cp.zeros([outlength], dtype= ak_array.dtype)

    # Initial value for the reduction
    # max value gets transformed to input_data.dtype automatically?
    max = cp.iinfo(index_dtype).max
    h_init = ak_array(max, max)

    # Perform the segmented reduce
    segmented_reduce(input_struct, _result, start_o, end_o, min_op, h_init, outlength)

    # TODO: here converts float to int too, fix this?
    _result = _result.view(index_dtype).reshape(-1, 2)
    _result = _result[:, 1]

    # pass the result outside the function
    result_v = result.view()
    result_v[...] = _result



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak

to_tensorflow = ak.operations.to_tensorflow
from_tensorflow = ak.operations.from_tensorflow

tf = pytest.importorskip("tensorflow")

a = np.arange(2 * 2 * 2, dtype=np.float64).reshape(2, 2, 2)
b = np.arange(2 * 2 * 2).reshape(2, 2, 2)

array = np.arange(2 * 3 * 5).reshape(2, 3, 5)
content2 = ak.contents.NumpyArray(array.reshape(-1))
inneroffsets = ak.index.Index64(np.array([0, 5, 10, 15, 20, 25, 30]))
outeroffsets = ak.index.Index64(np.array([0, 3, 6]))


def test_to_tensorflow():
    # a basic test for a 4 dimensional array
    array1 = ak.Array([a, b])
    i = 0
    for sub_array in [
        [[[0.0, 1.0], [2.0, 3.0]], [[4.0, 5.0], [6.0, 7.0]]],
        [[[0.0, 1.0], [2.0, 3.0]], [[4.0, 5.0], [6.0, 7.0]]],
    ]:
        assert to_tensorflow(array1)[i].numpy().tolist() == sub_array
        i += 1

    # test that the data types are remaining the same (float64 in this case)
    assert array1.layout.to_backend_array().dtype.name in str(
        to_tensorflow(array1).dtype
    )

    # try a listoffset array inside a listoffset array
    array2 = ak.contents.ListOffsetArray(
        outeroffsets, ak.contents.ListOffsetArray(inneroffsets, content2)
    )
    assert to_tensorflow(array2)[0].numpy().tolist() == [
        [0, 1, 2, 3, 4],
        [5, 6, 7, 8, 9],
        [10, 11, 12, 13, 14],
    ]
    assert to_tensorflow(array2)[1].numpy().tolist() == [
        [15, 16, 17, 18, 19],
        [20, 21, 22, 23, 24],
        [25, 26, 27, 28, 29],
    ]

    # try just a python list
    array3 = [3, 1, 4, 1, 9, 2, 6]
    assert to_tensorflow(array3).numpy().tolist() == [3, 1, 4, 1, 9, 2, 6]


array1 = tf.constant([[1.0, -1.0], [1.0, -1.0]], dtype=tf.float32)
array2 = tf.constant(np.array([[1, 2, 3], [4, 5, 6]]))


def test_from_tensorflow():
    # Awkward.to_list() == Tensor.numpy().tolist()
    assert from_tensorflow(array1).to_list() == array1.numpy().tolist()

    assert from_tensorflow(array2).to_list() == [[1, 2, 3], [4, 5, 6]]

    # test that the data types are remaining the same (int64 in this case)
    assert from_tensorflow(array1).layout.dtype.name in str(array1.dtype)

    # test that the data types are remaining the same (float32 in this case)
    assert from_tensorflow(array2).layout.dtype.name in str(array2.dtype)



# AUTO GENERATED ON 2026-02-08 AT 18:23:02
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import cupy.testing as cpt
import numpy as np
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cudaawkward_UnionArray_filltags_to8_from8_1():
    totags = cupy.array([123, 123, 123, 123, 123, 123], dtype=cupy.int8)
    totagsoffset = 3
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    base = 3
    funcC = cupy_backend['awkward_UnionArray_filltags', cupy.int8, cupy.int8]
    funcC(totags, totagsoffset, fromtags, length, base)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totags = [123, 123, 123, 3, 3, 3]
    cpt.assert_allclose(totags[:len(pytest_totags)], cupy.array(pytest_totags))

def test_cudaawkward_UnionArray_filltags_to8_from8_2():
    totags = cupy.array([123, 123, 123, 123, 123, 123], dtype=cupy.int8)
    totagsoffset = 3
    fromtags = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    base = 3
    funcC = cupy_backend['awkward_UnionArray_filltags', cupy.int8, cupy.int8]
    funcC(totags, totagsoffset, fromtags, length, base)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totags = [123, 123, 123, 4, 4, 4]
    cpt.assert_allclose(totags[:len(pytest_totags)], cupy.array(pytest_totags))

def test_cudaawkward_UnionArray_filltags_to8_from8_3():
    totags = cupy.array([123, 123, 123, 123, 123, 123], dtype=cupy.int8)
    totagsoffset = 3
    fromtags = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    base = 3
    funcC = cupy_backend['awkward_UnionArray_filltags', cupy.int8, cupy.int8]
    funcC(totags, totagsoffset, fromtags, length, base)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totags = [123, 123, 123, 4, 4, 4]
    cpt.assert_allclose(totags[:len(pytest_totags)], cupy.array(pytest_totags))

def test_cudaawkward_UnionArray_filltags_to8_from8_4():
    totags = cupy.array([123, 123, 123, 123, 123, 123], dtype=cupy.int8)
    totagsoffset = 3
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    base = 3
    funcC = cupy_backend['awkward_UnionArray_filltags', cupy.int8, cupy.int8]
    funcC(totags, totagsoffset, fromtags, length, base)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totags = [123, 123, 123, 3, 3, 3]
    cpt.assert_allclose(totags[:len(pytest_totags)], cupy.array(pytest_totags))

def test_cudaawkward_UnionArray_filltags_to8_from8_5():
    totags = cupy.array([123, 123, 123, 123, 123, 123], dtype=cupy.int8)
    totagsoffset = 3
    fromtags = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    base = 3
    funcC = cupy_backend['awkward_UnionArray_filltags', cupy.int8, cupy.int8]
    funcC(totags, totagsoffset, fromtags, length, base)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_totags = [123, 123, 123, 3, 3, 3]
    cpt.assert_allclose(totags[:len(pytest_totags)], cupy.array(pytest_totags))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest

import awkward as ak

pandas = pytest.importorskip("pandas")


def test_masked_string_array_with_option():
    """Test that string arrays with None values convert properly to dataframe."""
    # Create a string array with None values
    # This will create masked arrays internally during conversion
    array_str = ak.Array([["a", None, "c"], ["d", "e", None]])

    # Convert to dataframe - should not raise an error about dtype width
    df = ak.operations.to_dataframe(array_str)

    # Verify the conversion worked and "nan" appears for None values
    assert df["values"].values[0] == "a"
    assert df["values"].values[1] == "nan"
    assert df["values"].values[2] == "c"
    assert df["values"].values[3] == "d"
    assert df["values"].values[4] == "e"
    assert df["values"].values[5] == "nan"


def test_masked_bytestring_array_with_option():
    """Test that bytestring arrays with None values convert properly to dataframe."""
    # Create a bytestring array with None values
    array_bytes = ak.Array([[b"x", None, b"z"], [b"a", b"b", None]])

    # Convert to dataframe - should not raise an error about dtype width
    df = ak.operations.to_dataframe(array_bytes)

    # Verify the conversion worked and b"nan" appears for None values
    assert df["values"].values[0] == b"x"
    assert df["values"].values[1] == b"nan"
    assert df["values"].values[2] == b"z"
    assert df["values"].values[3] == b"a"
    assert df["values"].values[4] == b"b"
    assert df["values"].values[5] == b"nan"


def test_union_with_narrow_strings():
    """Test union arrays with narrow strings and None values."""
    # Test case from the existing test_union_to_record
    unionarray = ak.Array(
        [{"x": "a", "y": 1.1}, {"y": 2.2, "z": "b"}, {"x": "c", "y": 3.3}]
    )

    df = ak.operations.to_dataframe(unionarray)

    # Verify conversion works without dtype errors
    assert len(df) == 3
    # The x column should have "nan" for missing values
    assert df["x"].values[1] == "nan"


def test_single_char_strings_with_none():
    """Test very short strings with None values."""
    # Single character strings with None values
    # This is the edge case that needs dtype resizing
    array = ak.Array([["a", "b"], [None, "c"], ["d", None]])

    df = ak.operations.to_dataframe(array)

    # Verify all values are present
    values = df["values"].values.tolist()
    assert "a" in values
    assert "b" in values
    assert "c" in values
    assert "d" in values
    # Check that "nan" string appears for None values
    assert values.count("nan") == 2


def test_single_byte_bytestrings_with_none():
    """Test very short bytestrings with None values."""
    # Single byte strings with None values
    array = ak.Array([[b"a", b"b"], [None, b"c"], [b"d", None]])

    df = ak.operations.to_dataframe(array)

    # Verify all values are present
    values = df["values"].values.tolist()
    assert b"a" in values
    assert b"b" in values
    assert b"c" in values
    assert b"d" in values
    # Check that b"nan" appears for None values
    assert values.count(b"nan") == 2


def test_two_char_strings_with_none():
    """Test two-character strings with None values (edge case)."""
    # Two character strings - exactly the edge case where "nan" (3 chars) won't fit
    array = ak.Array([["ab", None], [None, "cd"]])

    df = ak.operations.to_dataframe(array)

    values = df["values"].values.tolist()
    assert "ab" in values
    assert "cd" in values
    # "nan" should appear (3 characters)
    assert values.count("nan") == 2



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE
from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext

__all__ = ("materialize",)


@high_level_function()
def materialize(
    array,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        array : Array-like data (either an #ak.Array or an #ak.contents.Content).
            An array that may contain virtual buffers to be materialized.
        highlevel (bool, default is True): If True, return an #ak.Array;
            otherwise, return a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Traverses the input array and materializes any virtual buffers.
    If the input array is not an #ak.Array or an #ak.contents.Content,
    an error will be raised.
    The buffers of the returned array are no longer `VirtualNDArray` objects even if there were any.
    They will become one of `numpy.ndarray`, `cupy.ndarray`, or `jax.numpy.ndarray` objects,
    depending on the array's backend.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, highlevel, behavior, attrs)


def _impl(array, highlevel, behavior, attrs):
    if not isinstance(array, (ak.highlevel.Array, ak.contents.Content)):
        raise TypeError(
            f"Only an ak.Array (or low-level equivalent) should be passed into ak.materialize. Received {type(array)}"
        )

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(
            array,
            allow_record=False,
            primitive_policy="error",
            string_policy="error",
            use_from_iter=False,
        )
    out = layout.materialize()
    return ctx.wrap(out, highlevel=highlevel)



from __future__ import annotations

import pickle

import pytest

import awkward as ak


def test_to_from_buffers():
    a = ak.Array({"a": [1, 2, 3]})[["a", "a"]]
    b = ak.from_buffers(*ak.to_buffers(a))

    assert ak.to_list(b) == ak.to_list(a)
    assert ak.type(b) == ak.type(a)


def test_pickle():
    a = ak.Array({"a": [1, 2, 3]})[["a", "a"]]
    b = pickle.loads(pickle.dumps(a))

    assert ak.to_list(b) == ak.to_list(a)
    assert ak.type(b) == ak.type(a)


def test_arrow():
    pytest.importorskip("pyarrow")

    a = ak.Array({"a": [1, 2, 3]})[["a", "a"]]
    b = ak.from_arrow(ak.to_arrow(a))

    assert ak.to_list(b) == ak.to_list(a)
    assert ak.type(b) == ak.type(a)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest

import awkward as ak

pa = pytest.importorskip("pyarrow")


def test_strings():
    array = pa.chunked_array([["foo", "bar"], ["blah", "bleh"]])
    ak_array = ak.from_arrow(array)
    assert ak_array.type == ak.types.ArrayType(
        ak.types.ListType(
            ak.types.NumpyType("uint8", parameters={"__array__": "char"}),
            parameters={"__array__": "string"},
        ),
        4,
    )


def test_strings_option():
    array = pa.chunked_array([["foo", "bar"], ["blah", "bleh", None]])
    ak_array = ak.from_arrow(array)
    assert ak_array.type == ak.types.ArrayType(
        ak.types.OptionType(
            ak.types.ListType(
                ak.types.NumpyType("uint8", parameters={"__array__": "char"}),
                parameters={"__array__": "string"},
            )
        ),
        5,
    )


def test_numbers():
    array = pa.chunked_array([[1, 2, 3], [4, 5]])
    ak_array = ak.from_arrow(array)
    assert ak_array.type == ak.types.ArrayType(ak.types.NumpyType("int64"), 5)


def test_numbers_option():
    array = pa.chunked_array([[1, 2, 3], [4, 5, None]])
    ak_array = ak.from_arrow(array)
    assert ak_array.type == ak.types.ArrayType(
        ak.types.OptionType(ak.types.NumpyType("int64")), 6
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def test_select_columns():
    (records, records_tuple) = (
        ak.forms.RecordForm(
            [
                ak.forms.NumpyForm("int64"),
                ak.forms.NumpyForm("int64"),
            ],
            None if is_tuple else ["x", "y"],
        )
        for is_tuple in (False, True)
    )

    assert not records.is_tuple
    assert records_tuple.is_tuple

    assert not records.select_columns("*").is_tuple
    assert records_tuple.select_columns("*").is_tuple



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE
from __future__ import annotations

import pytest

import awkward as ak

jax = pytest.importorskip("jax")


def test():
    ak.jax.register_and_check()

    jets = ak.Array(
        [
            [
                {"pt": 1.0, "eta": 1.1, "phi": 0.1, "mass": 0.01},
                {"pt": 2, "eta": 2.2, "phi": 0.2, "mass": 0.02},
            ],
            [
                {"pt": 4.0, "eta": 4.4, "phi": 0.4, "mass": 0.04},
                {"pt": 5.0, "eta": 5.5, "phi": 0.5, "mass": 0.05},
                {"pt": 6.0, "eta": 6.6, "phi": 0.6, "mass": 0.06},
            ],
        ],
        backend="jax",
    )

    def correct_jets(jets, alpha):
        new_pt = jets["pt"] + 25.0 * alpha
        jets["pt"] = new_pt
        return ak.sum(jets["pt"])

    val, grad = jax.value_and_grad(correct_jets, argnums=1)(jets, 0.1)

    assert val == 30.5
    assert grad == 125.0



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE
# ruff: noqa: E402

from __future__ import annotations

import io
import os

import numpy as np
import pytest

pa = pytest.importorskip("pyarrow")
pq = pytest.importorskip("pyarrow.parquet")

import awkward as ak
from awkward._connect.pyarrow.table_conv import (
    AWKWARD_INFO_KEY,
    array_with_replacement_type,
    awkward_arrow_field_to_native,
    collect_ak_arr_type_metadata,
    convert_awkward_arrow_table_to_native,
    convert_native_arrow_table_to_awkward,
    native_arrow_field_to_akarraytype,
)
from awkward.operations import to_list

nested_ints = ak.Array([[[[[1, 2, 3], [], [4, 5]] * 5] * 3] * 2])

struct_array = ak.Array(
    [
        {"lbl": "item 1", "idx": 11, "ids": [1, 2, 3]},
        {"lbl": "item 2", "idx": 12, "ids": [51, 52]},
        {"lbl": "item 3", "idx": 13, "ids": [61, 62, 63, 64]},
    ],
    with_name="my_items",
)

with_nulls = ak.Array([[11, None], [23], None])

option_types = ak.Array(
    [[1, 2], [[10, 11, None], None, [3, 4, 5], ["one", "two"]], None]
)

(indexed, indexed_tuple) = (
    # Unique for this test, contains nested types within the
    ak.contents.IndexedArray(
        ak.index.Index64(np.array([2, 1, 0], dtype=np.uint64)),
        ak.contents.RecordArray(
            [
                ak.contents.ByteMaskedArray(
                    ak.index.Index8(np.array([False, True, False]).view(np.int8)),
                    ak.contents.NumpyArray(np.array([1.1, 2.2, 3.3])),
                    valid_when=False,
                ),
                ak.contents.UnmaskedArray(
                    ak.contents.ListOffsetArray(
                        ak.index.Index32(np.array([0, 3, 3, 5], dtype=np.int32)),
                        ak.contents.NumpyArray(
                            np.array([1.1, 2.2, 3.3, 4.4, 5.5]),
                        ),
                    ),
                ),
            ],
            None if is_tuple else ["x", "y"],
        ),
        parameters={"__array__": "categorical", "foo": "thap"},
    )
    for is_tuple in (False, True)
)

(records, records_tuple) = (
    ak.contents.RecordArray(
        [
            ak.contents.NumpyArray(
                np.array([1.1, 2.2, 3.3]), parameters={"foo": "inner1"}
            ),
            ak.contents.ListOffsetArray(
                ak.index.Index32(np.array([0, 3, 3, 5], dtype=np.int32)),
                ak.contents.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5]), parameters={"foo": "inner2"}
                ),
            ),
        ],
        None if is_tuple else ["x", "y"],
        parameters={"foo": "outer"},
    )
    for is_tuple in (False, True)
)


@pytest.mark.parametrize(
    "akarray, as_dict",
    [
        (nested_ints, False),
        (struct_array, False),
        (with_nulls, False),
        (option_types, False),
        (indexed, True),
        (indexed_tuple, True),
        (records, True),
        (records_tuple, True),
        (indexed, False),
        (indexed_tuple, False),
        (records, False),
        (records_tuple, False),
    ],
)
def test_array_conversions(akarray, as_dict):
    arrow_natv = ak.to_arrow(
        akarray, categorical_as_dictionary=as_dict, extensionarray=False
    )
    arrow_extn = ak.to_arrow(
        akarray, categorical_as_dictionary=as_dict, extensionarray=True
    )
    extn_field = pa.field("test_field", arrow_extn.type)

    # Convert to native pyarrow types:
    metadata = collect_ak_arr_type_metadata(extn_field)
    conv_natv_field = awkward_arrow_field_to_native(extn_field)
    assert conv_natv_field.type == arrow_natv.type
    as_natv = array_with_replacement_type(arrow_extn, conv_natv_field.type)
    assert as_natv == arrow_natv

    # Complete a round-trip, back to AwkwardArrowArray
    conv_extn_field = native_arrow_field_to_akarraytype(conv_natv_field, metadata)
    assert conv_extn_field.type == arrow_extn.type
    assert (
        conv_extn_field.type._metadata_as_dict() == arrow_extn.type._metadata_as_dict()
    )
    conv_metadata = collect_ak_arr_type_metadata(conv_extn_field)
    assert conv_metadata == metadata
    # assert_equal_arrowextntypes(conv_extn_field.type, arrow_extn.type)
    as_extn = array_with_replacement_type(as_natv, conv_extn_field.type)
    assert as_extn == arrow_extn
    assert as_extn.type._metadata_as_dict() == arrow_extn.type._metadata_as_dict()

    # And back to Awkward array
    rt_array = ak.from_arrow(as_extn, highlevel=True)
    assert to_list(rt_array) == to_list(akarray)

    # Deeper test of types
    akarray_high = ak.Array(akarray)
    if akarray_high.type.content.parameters.get("__categorical__", False) == as_dict:
        # as_dict is supposed to go hand-in-hand with __categorical__: True, and if it
        #  does not, we do not round-trip perfectly. So only test when this is set correctly.
        assert rt_array.type == akarray_high.type

        ak_type_str_orig = io.StringIO()
        ak_type_str_rtrp = io.StringIO()
        akarray_high.type.show(stream=ak_type_str_orig)
        rt_array.type.show(stream=ak_type_str_rtrp)
        if ak_type_str_orig.getvalue() != ak_type_str_rtrp.getvalue():
            print("  Original type:", ak_type_str_orig.getvalue())
            print("  Rnd-trip type:", ak_type_str_rtrp.getvalue())
        assert ak_type_str_orig.getvalue() == ak_type_str_rtrp.getvalue()


def test_table_conversion():
    ak_tbl_like = ak.Array(
        {
            "struct_array": struct_array,
            "with_nulls": with_nulls,
            "option_types": option_types,
            "indexed": indexed,
        }
    )
    natv_tbl = ak.to_arrow_table(
        ak_tbl_like, categorical_as_dictionary=True, extensionarray=False
    )
    extn_tbl = ak.to_arrow_table(
        ak_tbl_like, categorical_as_dictionary=True, extensionarray=True
    )

    conv_natv_tbl = convert_awkward_arrow_table_to_native(extn_tbl)
    assert (
        conv_natv_tbl.schema == natv_tbl.schema
    )  # This comparison does not include metadata
    assert conv_natv_tbl == natv_tbl
    assert AWKWARD_INFO_KEY in conv_natv_tbl.schema.metadata
    assert len(conv_natv_tbl["struct_array"].chunks[0].field(2).buffers()) == 4
    # This verifies that our conversions are zero-copy for the table data
    assert (
        extn_tbl["struct_array"].chunks[0].storage.field(2).buffers()[3].address
        == conv_natv_tbl["struct_array"].chunks[0].field(2).buffers()[3].address
    )

    conv_extn_tbl = convert_native_arrow_table_to_awkward(conv_natv_tbl)
    assert conv_extn_tbl.schema == extn_tbl.schema
    assert conv_extn_tbl == extn_tbl  # Full Round-trip verification
    assert AWKWARD_INFO_KEY not in conv_extn_tbl.schema.metadata
    assert (
        extn_tbl["struct_array"].chunks[0].storage.field(2).buffers()[3].address
        == conv_extn_tbl["struct_array"].chunks[0].storage.field(2).buffers()[3].address
    )


def test_selective_parquet(tmp_path):
    filename = os.path.join(tmp_path, "whatever.parquet")
    ak_tbl = ak.Array(
        {
            "with_nulls": with_nulls,
            "struct_array": struct_array,
            "indexed": indexed,
        }
    )
    ak.to_parquet(ak_tbl, filename)
    tbl_tr = ak.from_parquet(filename, columns=["struct_array", "indexed"])
    assert to_list(tbl_tr["struct_array"]) == to_list(ak_tbl["struct_array"])


@pytest.mark.parametrize("doit", [False, True])
def test_empty(tmp_path, doit):
    filename = os.path.join(tmp_path, "whatever.parquet")

    ak.to_parquet(ak.Array([{"x": 1, "y": 1.1}])[0:0], filename, extensionarray=doit)

    assert str(ak.from_parquet(filename).type) == "0 * {x: int64, y: float64}"



from __future__ import annotations

import json
import re
import threading
from dataclasses import dataclass

import awkward._typing as tp
from awkward._layout import _neg2pos_axis
from awkward._regularize import is_integer

# axis names are hashables, mostly strings,
# except for integers, which are reserved for positional axis.
AxisName: tp.TypeAlias = tp.Hashable

# e.g.: {"x": 0, "y": 1, "z": 2}
AxisMapping: tp.TypeAlias = tp.Mapping[AxisName, int]

# e.g.: ("x", "y", None) where None is a wildcard
AxisTuple: tp.TypeAlias = tp.Tuple[AxisName, ...]


NAMED_AXIS_KEY: tp.Literal["__named_axis__"] = (
    "__named_axis__"  # reserved for named axis
)


class _NamedAxisLocal(threading.local):
    def __init__(self):
        self.mapping = {}


NamedAxis = _NamedAxisLocal()


def _prettify_named_axes(
    named_axis: AxisMapping,
    delimiter: str = ", ",
    maxlen: None | int = None,
) -> str:
    """
    This function takes a named axis mapping and returns a string representation of the mapping.
    The axis names are sorted in ascending order of their corresponding integer values.
    If the axis name is a valid Python identifier, it is represented as is.
    Otherwise, it is represented as a JSON string.

    Args:
        named_axis (AxisMapping): The named axis mapping to prettify.
        delimiter (str, optional): The delimiter to use between items in the output string. Defaults to ", ".
        maxlen (None | int, optional): The maximum length of the output string. If the string exceeds this length, it is truncated and ends with "...". Defaults to None.

    Returns:
        str: The prettified string representation of the named axis mapping.

    Examples:
        >>> _prettify_named_axes({"x": 0, "y": 1, "z": 2})
        'x:0, y:1, z:2'
        >>> _prettify_named_axes({"x": 0, "y": 1, "$": 2})
        'x:0, y:1, "$":2'
        >>> _prettify_named_axes({"x": 0, "y": 1, "z": 2}, delimiter="; ")
        'x:0; y:1; z:2'
        >>> _prettify_named_axes({"foo": 0, "bar": 1, "baz": 2}, maxlen=17)
        'foo:0, bar:1, ...'
    """

    def _prettify(ax: AxisName) -> str:
        repr_ax = str(ax)
        if re.match("[A-Za-z_][A-Za-z_0-9]*", repr_ax):
            return repr_ax
        return json.dumps(repr_ax)

    sorted_named_axis = sorted(named_axis.items(), key=lambda x: x[1])
    items = [
        f"{_prettify(named_ax)}:{pos_ax}" for named_ax, pos_ax in sorted_named_axis
    ]
    if maxlen is not None:
        if len(delimiter.join(items)) > maxlen:
            while (
                len(delimiter.join(items)) > maxlen - len(delimiter + "...")
            ) and items:
                items.pop(-1)
            items.append("...")
    return delimiter.join(items)


def _get_named_axis(ctx: tp.Any, allow_any: bool = False) -> AxisMapping:
    """
    Retrieves the named axis from the provided context.

    Args:
        ctx (Any): The context from which the named axis is to be retrieved.

    Returns:
        AxisMapping: The named axis retrieved from the context. If the context does not include a named axis,
            an empty dictionary is returned.

    Examples:
        >>> _get_named_axis(ak.Array([1, 2, 3], named_axis={"x": 0}))
        {"x": 0}
        >>> _get_named_axis(np.array([1, 2, 3]))
        {}
        >>> _get_named_axis({NAMED_AXIS_KEY: {"x": 0, "y": 1, "z": 2}})
        {"x": 0, "y": 1, "z": 2}
        >>> _get_named_axis({"other_key": "other_value"})
        {}
    """
    from awkward._layout import HighLevelContext
    from awkward.highlevel import Array, Record

    if hasattr(ctx, "attrs") and (
        isinstance(ctx, (HighLevelContext, Array, Record)) or allow_any
    ):
        return _get_named_axis(ctx.attrs, allow_any=True)
    elif allow_any and isinstance(ctx, tp.Mapping) and NAMED_AXIS_KEY in ctx:
        return dict(ctx[NAMED_AXIS_KEY])
    else:
        return {}


def _make_positional_axis_tuple(n: int) -> tuple[int, ...]:
    """
    Generates a positional axis tuple of length n.

    Args:
        n (int): The length of the positional axis tuple to generate.

    Returns:
        tuple[int, ...]: The generated positional axis tuple.

    Examples:
        >>> _make_positional_axis_tuple(3)
        (0, 1, 2)
    """
    return tuple(range(n))


def _is_valid_named_axis(axis: AxisName) -> bool:
    """
    Checks if the given axis is a valid named axis. A valid named axis is a hashable object that is not an integer or None. Currently it is restricted to strings.

    Args:
        axis (AxisName): The axis to check.

    Returns:
        bool: True if the axis is a valid named axis, False otherwise.

    Examples:
        >>> _is_valid_named_axis("x")
        True
        >>> _is_valid_named_axis(1)
        False
    """
    return (
        # axis must be hashable
        isinstance(axis, AxisName)
        # ... but not an integer, otherwise we would confuse it with positional axis
        and not is_integer(axis)
        # we also prohibit None, which is reserved for wildcard
        and axis is not None
        # Let's only allow strings for now, in the future we can open up to more types
        # by removing the isinstance(axis, str) check.
        and isinstance(axis, str)
    )


def _check_valid_axis(axis: AxisName) -> AxisName:
    """
    Checks if the given axis is a valid named axis. If not, raises a ValueError.

    Args:
        axis (AxisName): The axis to check.

    Returns:
        AxisName: The axis if it is a valid named axis.

    Raises:
        ValueError: If the axis is not a valid named axis.

    Examples:
        >>> _check_valid_axis("x")
        "x"
        >>> _check_valid_axis(1)
        Traceback (most recent call last):
        ...
        ValueError: Axis names must be hashable and not int, got 1 [type(axis)=<class 'int'>]
    """
    if not _is_valid_named_axis(axis):
        raise ValueError(
            f"Axis names must be hashable and not int, got {axis!r} [{type(axis)=}]"
        )
    return axis


def _check_valid_named_axis_mapping(named_axis: AxisMapping) -> AxisMapping:
    """
    Checks if the given named axis mapping is valid. A valid named axis mapping is a dictionary where the keys are valid named axes
    (hashable objects that are not integers) and the values are integers.

    Args:
        named_axis (AxisMapping): The named axis mapping to check.

    Raises:
        ValueError: If any of the keys in the named axis mapping is not a valid named axis or if any of the values is not an integer.

    Examples:
        >>> _check_valid_named_axis_mapping({"x": 0, "y": 1, "z": 2})  # No exception is raised
        >>> _check_valid_named_axis_mapping({"x": 0, "y": 1, "z": "2"})
        Traceback (most recent call last):
        ...
        ValueError: Named axis mapping values must be integers, got '2' [type(axis)=<class 'str'>]
        >>> _check_valid_named_axis_mapping({"x": 0, 1: 1, "z": 2})
        Traceback (most recent call last):
        ...
        ValueError: Axis names must be hashable and not int, got 1 [type(axis)=<class 'int'>]
    """
    for name, axis in named_axis.items():
        _check_valid_axis(name)
        if not is_integer(axis):
            raise ValueError(
                f"Named axis mapping values must be integers, got {axis!r} [{type(axis)=}]"
            )
    return named_axis


def _axis_tuple_to_mapping(axis_tuple: AxisTuple) -> AxisMapping:
    """
    Converts a tuple of axis names to a dictionary mapping axis names to their positions.

    Args:
        axis_tuple (AxisTuple): A tuple of axis names. Can include None as a wildcard.

    Returns:
        AxisMapping: A dictionary mapping axis names to their positions.

    Examples:
        >>> _axis_tuple_to_mapping(("x", None, "y"))
        {"x": 0, "y": 2}
    """
    return {axis: i for i, axis in enumerate(axis_tuple) if axis is not None}


def _prepare_named_axis_for_attrs(
    named_axis: AxisMapping | AxisTuple,
    ndim: int,
) -> AxisMapping:
    """
    Prepares the named axis for attribute assignment.

    This function takes a named axis, which can either be a mapping or a tuple, and returns a dictionary mapping axis names to their positions.
    The function checks if the named axis is valid and if the positional axes match the number of dimensions. If not, an error is raised.

    Args:
        named_axis (AxisMapping | AxisTuple): The named axis to prepare. Can either be a mapping or a tuple.
        ndim (int): The number of dimensions.

    Returns:
        AxisMapping: The prepared named axis.

    Raises:
        TypeError: If the named axis is not a mapping or a tuple.
        ValueError: If the named axes do not point to positional axes matching the number of dimensions.

    Examples:
        >>> _prepare_named_axis_for_attrs({"x": 0, "y": 1, "z": 2}, 3)
        {"x": 0, "y": 1, "z": 2}
        >>> _prepare_named_axis_for_attrs(("x", "y", "z"), 3)
        {"x": 0, "y": 1, "z": 2}
        >>> _prepare_named_axis_for_attrs({"x": 0, "y": 1, "z": 2}, 2)
        Traceback (most recent call last):
        ...
        ValueError: Named axes must point to positional axes matching 2 dimensions, got named_axis={"x": 0, "y": 1, "z": 2}, ndim=2
    """
    if isinstance(named_axis, tuple):
        _named_axis = _axis_tuple_to_mapping(named_axis)
    elif isinstance(named_axis, dict):
        _named_axis = named_axis
    else:
        raise TypeError(
            f"named_axis must be a mapping or a tuple, got {named_axis=} [{type(named_axis)=}]"
        )
    _check_valid_named_axis_mapping(_named_axis)
    pos_axes = set(_named_axis.values())
    if max(pos_axes, default=0) >= ndim or min(pos_axes, default=0) < -ndim:
        raise ValueError(
            f"Named axes must point to positional axes matching {ndim} dimensions, got {named_axis=}, {ndim=}"
        )
    return _named_axis


def _make_named_int_class(name: tp.Any) -> type[int]:
    class NamedInt(int):
        def __repr__(self):
            value_repr = super().__repr__()
            return f"{name!r} (named axis) -> {value_repr} (pos. axis)"

        __str__ = __repr__

    return NamedInt


def _named_axis_to_positional_axis(
    named_axis: AxisMapping,
    axis: AxisName,
) -> int | None:
    """
    Converts a single named axis to a positional axis.

    Args:
        axis (AxisName): The named axis to convert.
        named_axis (AxisMapping): The mapping from named axes to positional axes.

    Returns:
        int | None: The positional axis corresponding to the given named axis. If the named axis is not found in the mapping, returns None.

    Raises:
        ValueError: If the named axis is not found in the named axis mapping.

    Examples:
        >>> _named_axis_to_positional_axis({"x": 0, "y": 1, "z": 2}, "x")
        0
    """
    if _is_valid_named_axis(axis):
        if axis not in named_axis:
            raise ValueError(f"{axis=} not found in {named_axis=} mapping.")

        # we wrap it to preserve the original name in its __repr__ and __str__
        # in order to properly display it in error messages. This is useful for cases
        # where the positional axis is pointing to a non-existing axis. The error message
        # will then show the original (named) axis together with the positional axis.
        cls = _make_named_int_class(axis)
        return cls(named_axis[axis])

    if is_integer(axis):
        # TODO: is_integer is an external  helper function that doesn't specify types
        return int(tp.cast(tp.Any, axis))
    elif axis is None:
        return None
    else:
        raise ValueError(f"Invalid {axis=} [{type(axis)=}]")


# These are the strategies to handle named axis for the
# output array when performing operations along an axis.
# See studies/named_axis.md#named-axis-in-high-level-functions and
# https://pytorch.org/docs/stable/name_inference.html.
#
# The possible strategies are:
# - "keep all" (_keep_named_axis(..., None)): Keep all named axes in the output array, e.g.: `ak.drop_none`
# - "keep one" (_keep_named_axis(..., int)): Keep one named axes in the output array, e.g.: `ak.firsts`
# - "keep up to" (_keep_named_axis_up_to(..., int)): Keep all named axes up to a certain positional axis in the output array, e.g.: `ak.local_index`
# - "remove all" (_remove_all_named_axis): Removes all named axis, e.g.: `ak.categories`
# - "remove one" (_remove_named_axis): Remove the named axis from the output array, e.g.: `ak.sum`
# - "add one" (_add_named_axis): Add a new named axis to the output array, e.g.: `ak.concatenate`
# - "unify" (_unify_named_axis): Unify the named axis in the output array given two input arrays, e.g.: `ak.broadcast_arrays`


def _keep_named_axis(
    named_axis: AxisMapping,
    axis: int | None = None,
) -> AxisMapping:
    """
    Determines the new named axis after keeping the specified axis. This function is useful when an operation
    is applied that retains only one axis.

    Args:
        named_axis (AxisMapping): The current named axis.
        axis (int | None, optional): The index of the axis to keep. If None, all axes are kept. Default is None.

    Returns:
        AxisMapping: The new named axis after keeping the specified axis.

    Examples:
        >>> _keep_named_axis({"x": 0, "y": 1, "z": 2}, 1)
        {"y": 0}
        >>> _keep_named_axis({"x": 0, "y": 1, "z": 2}, None)
        {"x": 0, "y": 1, "z": 2}
    """
    if axis is None:
        return named_axis
    return {k: 0 for k, v in named_axis.items() if v == axis}


def _keep_named_axis_up_to(
    named_axis: AxisMapping,
    axis: int,
    total: int,
) -> AxisMapping:
    """
    Determines the new named axis after keeping all axes up to the specified axis. This function is useful when an operation
    is applied that retains all axes up to a certain axis.

    Args:
        named_axis (AxisMapping): The current named axis.
        axis (int): The index of the axis up to which to keep.
        total (int): The total number of axes.

    Returns:
        AxisMapping: The new named axis after keeping all axes up to the specified axis.

    Examples:
        >>> _keep_named_axis_up_to({"x": 0, "y": 1, "z": 2}, 1, 3)
        {"x": 0, "y": 1}
        >>> _keep_named_axis_up_to({"x": 0, "y": 1, "z": 2}, -1, 3)
        {"x": 0, "y": 1, "z": 2}
        >>> _keep_named_axis_up_to({"x": 0, "y": 1, "z": 2}, 0, 3)
        {"x": 0}
    """
    axis = _neg2pos_axis(axis, total)
    out = {}
    for k, v in named_axis.items():
        if v >= 0 and v <= axis:
            out[k] = v
        elif v < 0 and v >= -axis - 1:
            out[k] = v
    return out


def _remove_all_named_axis(
    named_axis: AxisMapping,
) -> AxisMapping:
    """
    Returns an empty named axis mapping after removing all axes from the given named axis mapping.
    This function is typically used when an operation that eliminates all axes is applied.

    Args:
        named_axis (AxisMapping): The current named axis mapping.

    Returns:
        AxisMapping: An empty named axis mapping.

    Examples:
        >>> _remove_all_named_axis({"x": 0, "y": 1, "z": 2})
        {}
    """
    return _remove_named_axis(named_axis=named_axis, axis=None)


def _remove_named_axis(
    named_axis: AxisMapping,
    axis: int | None = None,
    total: int | None = None,
) -> AxisMapping:
    """
    Determines the new named axis after removing the specified axis. This is useful, for example,
    when applying an operation that removes one axis.

    Args:
        named_axis (AxisMapping): The current named axis.
        axis (int | None, optional): The index of the axis to remove. If None, no axes are removed. Default is None.
        total (int | None, optional): The total number of axes. If None, it is calculated as the length of the named axis. Default is None.

    Returns:
        AxisMapping: The new named axis after removing the specified axis.

    Examples:
        >>> _remove_named_axis({"x": 0, "y": 1}, None)
        {}
        >>> _remove_named_axis({"x": 0, "y": 1}, 0)
        {"y": 0}
        >>> _remove_named_axis({"x": 0, "y": 1, "z": 2}, 1)
        {"x": 0, "z": 1}
        >>> _remove_named_axis({"x": 0, "y": 1, "z": -1}, 1)
        {"x": 0, "z": -1}
        >>> _remove_named_axis({"x": 0, "y": 1, "z": -3}, 1)
        {"x": 0, "z": -2}
    """
    if axis is None:
        return {}

    if total is None:
        total = len(named_axis)

    # remove the specified axis
    out = {
        ax: pos
        for ax, pos in named_axis.items()
        if _neg2pos_axis(pos, total) != _neg2pos_axis(axis, total)
    }

    return _adjust_pos_axis(out, axis, total, direction=-1)


def _adjust_pos_axis(
    named_axis: AxisMapping,
    axis: int,
    total: int,
    direction: int,
) -> AxisMapping:
    """
    Adjusts the positions of the axes in the named axis mapping after an axis has been removed or added.

    Args:
        named_axis (AxisMapping): The current named axis mapping.
        axis (int): The position of the removed/added axis.
        total (int): The total number of axes.
        direction (int): The direction of the adjustment. -1 means axis is removed; +1 means axis is added. Default is +1.

    Returns:
        AxisMapping: The adjusted named axis mapping.

    Examples:
        # axis=1 removed
        >>> _adjust_pos_axis({"x": 0, "z": 2}, 1, 3, -1)
        {"x": 0, "z": 1}
        # axis=1 added
        >>> _adjust_pos_axis({"x": 0, "z": 2}, 1, 3, +1)
        {"x": 0, "z": 3}
        # axis=1 removed
        >>> _adjust_pos_axis({"x": 0, "z": -1}, 1, 3, -1)
        {"x": 0, "z": -1}
        # axis=1 added
        >>> _adjust_pos_axis({"x": 0, "z": -1}, 1, 3, +1)
        {"x": 0, "z": -1}
    """
    assert direction in (-1, +1), f"Invalid direction: {direction}"

    def _adjust(pos: int, axis: int, direction: int) -> int:
        # positive axis
        if axis >= 0:
            # positive axis and position greater than or equal to the removed/added (positive) axis
            # -> change position by direction
            if pos >= axis:
                return pos + direction
            # positive axis and negative position
            # -> change position by direction
            elif pos < 0 and pos + total < axis:
                return pos - direction
            # positive axis and position smaller than the removed/added (positive) axis, but greater than 0
            # -> keep position
            else:
                return pos
        # negative axis
        else:
            # negative axis and position smaller than the removed/added (negative) axis
            # -> change position by inverse direction
            if pos <= axis:
                return pos - direction
            # negative axis and positive position
            # -> change position by inverse direction
            elif pos > 0 and pos > axis + total:
                return pos + direction
            # negative axis and position greater than the removed/added (negative) axis, but smaller than 0
            # -> keep position
            else:
                return pos

    return {k: _adjust(v, axis, direction) for k, v in named_axis.items()}


def _add_named_axis(
    named_axis: AxisMapping,
    axis: int,
    total: int | None = None,
) -> AxisMapping:
    """
    Adds a new axis to the named_axis at the specified position.

    Args:
        named_axis (AxisMapping): The current named axis mapping.
        axis (int): The position at which to add the new axis.
        total (int | None): The total number of axes.

    Returns:
        AxisMapping: The updated named axis mapping after adding the new axis.

    Examples:
        >>> _add_named_axis({"x": 0, "y": 1, "z": 2}, 0)
        {"x": 1, "y": 2, "z": 3}
        >>> _add_named_axis({"x": 0, "y": 1, "z": 2}, 1)
        {"x": 0, "y": 2, "z": 3}
    """
    if total is None:
        total = len(named_axis)

    return _adjust_pos_axis(named_axis, axis, total, direction=+1)


def _unify_named_axis(
    named_axis1: AxisMapping,
    named_axis2: AxisMapping,
) -> AxisMapping:
    """
    Unifies two named axes into a single named axis. The function iterates over all positional axes present in either of the input named axes.
    For each positional axis, it checks the corresponding axis names in both input named axes. If the axis names are the same or if one of them is None,
    the unified axis will be the non-None axis. If the axis names are different and neither of them is None, a ValueError is raised.

    Args:
        named_axis1 (AxisMapping): The first named axis to unify.
        named_axis2 (AxisMapping): The second named axis to unify.

    Returns:
        AxisMapping: The unified named axis.

    Raises:
        ValueError: If the axes are different and neither of them is None.

    Examples:
        >>> _unify_named_axis({"x": 0, "y": 1, "z": 2}, {"x": 0, "y": 1, "z": 2})
        {"x": 0, "y": 1, "z": 2}

        >>> _unify_named_axis({"x": 0, "y": 1}, {"x": 0, "y": 1, "z": 2})
        {"x": 0, "y": 1, "z": 2}

        >>> _unify_named_axis({"x": 0, "y": 1, "z": 2}, {"a": 0, "b": 1, "c": 2})
        Traceback (most recent call last):
        ...
        ValueError: The named axes are different. Got: 'x' and 'a' for positional axis 0

        >>> _unify_named_axis({"x": 0, "y": 1, "z": 2}, {"x": 0, "y": 1, "z": 3})
        {"x": 0, "y": 1, "z": 2}

        >>> _unify_named_axis({"x": 0, "y": 1, "z": 2}, {})
        {"x": 0, "y": 1, "z": 2}

        >>> _unify_named_axis({}, {"x": 0, "y": 1, "z": 2})
        {"x": 0, "y": 1, "z": 2}

        >>> _unify_named_axis({}, {})
        {}
    """

    def _get_axis_name(
        axis_mapping: AxisMapping, positional_axis: int
    ) -> AxisName | None:
        for name, position in axis_mapping.items():
            if position == positional_axis:
                return name
        return None

    unified_named_axis = {}
    all_positional_axes = set(named_axis1.values()) | set(named_axis2.values())
    for position in all_positional_axes:
        axis_name1 = _get_axis_name(named_axis1, position)
        axis_name2 = _get_axis_name(named_axis2, position)
        if axis_name1 is not None and axis_name2 is not None:
            if axis_name1 != axis_name2:
                raise ValueError(
                    f"The named axes are incompatible. Got: {axis_name1} and {axis_name2} for positional axis {position}"
                )
            unified_named_axis[axis_name1] = position
        elif axis_name1 is not None:  # axis_name2 is None
            unified_named_axis[axis_name1] = position
        elif axis_name2 is not None:  # axis_name1 is None
            unified_named_axis[axis_name2] = position
    return unified_named_axis


@dataclass
class NamedAxesWithDims:
    """
    A dataclass that stores the named axis and their corresponding dimensions.
    This is a helper class to store the named axis mapping and the number of
    dimensions of each named axis, which is useful for broadcasting.

    Attributes:
        named_axis (AxisMapping): The named axis mapping.
        ndims (Tuple[int]): The number of dimensions of the named axis.
    """

    named_axis: list[AxisMapping]
    ndims: list[int | None]

    def __post_init__(self):
        if len(self.named_axis) != len(self.ndims):
            raise ValueError(
                "The number of dimensions must match the number of named axis mappings."
            )

    def __iter__(self) -> tp.Iterator[tuple[AxisMapping, int | None]]:
        yield from zip(self.named_axis, self.ndims, strict=True)

    @classmethod
    def prepare_contexts(
        cls, arrays: tp.Sequence, unwrap_kwargs: dict | None = None
    ) -> tuple[dict, dict]:
        from awkward._layout import HighLevelContext
        from awkward._typetracer import MaybeNone

        # unwrap options
        arrays = [x.content if isinstance(x, MaybeNone) else x for x in arrays]

        _unwrap_kwargs = {"allow_unknown": True}
        if unwrap_kwargs is not None:
            _unwrap_kwargs.update(unwrap_kwargs)

        _named_axes = []
        _ndims = []
        for array in arrays:
            with HighLevelContext() as ctx:
                layout = ctx.unwrap(array, **_unwrap_kwargs)
            _named_axes.append(_get_named_axis(array))
            _ndims.append(getattr(layout, "minmax_depth", (None, None))[1])

        depth_context = {NAMED_AXIS_KEY: cls(_named_axes, _ndims)}
        lateral_context = {NAMED_AXIS_KEY: cls(_named_axes, _ndims)}
        return depth_context, lateral_context

    def __setitem__(
        self, index: int, named_axis_with_ndim: tuple[AxisMapping, int | None]
    ):
        named_axis, ndim = named_axis_with_ndim
        self.named_axis[index] = named_axis
        self.ndims[index] = ndim

    def __getitem__(self, index: int) -> tuple[AxisMapping, int | None]:
        return self.named_axis[index], self.ndims[index]

    def __len__(self) -> int:
        return len(self.named_axis)


# Define a type alias for a slice or int (can be a single axis or a sequence of axes)
AxisSlice: tp.TypeAlias = tuple | slice | int | tp.EllipsisType | None
NamedAxisSlice: tp.TypeAlias = tp.Dict[AxisName, AxisSlice]


def _normalize_named_slice(
    named_axis: AxisMapping,
    where: AxisSlice | NamedAxisSlice,
    total: int,
) -> AxisSlice:
    """
    Normalizes a named slice into a positional slice.

    This function takes a named slice (a dictionary mapping axis names to slices) and converts it into a positional slice
    (a tuple of slices). The positional slice can then be used to index an array.

    Args:
        named_axis (AxisMapping): The current named axis mapping.
        where (AxisSlice | NamedAxisSlice): The slice to normalize. Can be a single slice, a tuple of slices, or a dictionary mapping axis names to slices.
        total (int): The total number of axes.

    Returns:
        AxisSlice: The normalized slice.

    Raises:
        ValueError: If an invalid axis name is provided in the slice.

    Examples:
        >>> _normalize_named_slice({"x": 0, "y": 1, "z": 2}, {0: 0}, 3)
        (0, slice(None), slice(None))
        >>> _normalize_named_slice({"x": 0, "y": 1, "z": 2}, {-1: 0}, 3)
        (slice(None), slice(None), 0)
        >>> _normalize_named_slice({"x": 0, "y": 1, "z": 2}, {"x": 0}, 3)
        (0, slice(None), slice(None))
        >>> _normalize_named_slice({"x": 0, "y": 1, "z": 2}, {"x": 0, "y": 1}, 3)
        (0, 1, slice(None))
        >>> _normalize_named_slice({"x": 0, "y": 1, "z": 2}, {"x": 0, "y": 1, "z": ...}, 3)
        (0, 1, ...)
        >>> _normalize_named_slice({"x": 0, "y": 1, "z": 2}, {"x": 0, "y": 1, "z": slice(0, 1)}, 3)
        (0, 1, slice(0, 1))
        >>> _normalize_named_slice({"x": 0, "y": 1, "z": 2}, {"x": (0, 1)}, 3)
        ((0, 1), slice(None), slice(None))
        >>> _normalize_named_slice({"x": 0, "y": 1, "z": 2}, {"x": [0, 1]}, 3)
        ([0, 1], slice(None), slice(None))
    """
    if isinstance(where, dict):
        out_where: list[AxisSlice] = [slice(None)] * total
        for ax_name, ax_where in where.items():
            slice_ = ax_where if ax_where is not ... else slice(None)
            if is_integer(ax_name):
                # it's an integer, pyright doesn't get this
                idx = tp.cast(int, ax_name)
                out_where[idx] = slice_
            elif _is_valid_named_axis(ax_name):
                # it's an integer, pyright doesn't get this
                idx = tp.cast(int, _named_axis_to_positional_axis(named_axis, ax_name))
                out_where[idx] = slice_
            else:
                raise ValueError(f"Invalid axis name: {ax_name} in slice {where}")
        where = tuple(out_where)
    return where



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak

pyarrow = pytest.importorskip("pyarrow")
pyarrow_parquet = pytest.importorskip("pyarrow.parquet")


def test_parquet_string_roundtrip_with_string_to32_default(tmp_path):
    """Test that strings can be round-tripped through parquet with default string_to32=True.

    This test ensures that when string_to32=True (the default for parquet), strings are
    correctly serialized to Arrow's 32-bit string type and can be converted back to NumPy
    arrays without kernel lookup errors.

    Regression test for bug where int32 offsets from pyarrow.string() caused KeyError
    when converting to NumPy because UTF8 kernels only had int64 specializations.
    """
    # Create array with strings
    a = ak.Array([{"foo": "bar"}, {"foo": "baz"}])

    # Convert to numpy before serialization (should work with int64 offsets)
    b = np.asarray(a["foo"])
    assert b.dtype.kind == "U"
    assert list(b) == ["bar", "baz"]

    # Serialize to parquet with default string_to32=True
    ak.to_parquet(a, tmp_path / "test_string_to32.parquet")

    # Deserialize from parquet (creates int32 offsets for strings)
    x = ak.from_parquet(tmp_path / "test_string_to32.parquet")

    # This should work now with int32 offset kernel specializations
    y = np.asarray(x["foo"])
    assert y.dtype.kind == "U"
    assert list(y) == ["bar", "baz"]


def test_parquet_string_roundtrip_explicit_string_to32_true(tmp_path):
    """Test parquet roundtrip with explicitly setting string_to32=True."""
    a = ak.Array([{"name": "Alice"}, {"name": "Bob"}, {"name": "Charlie"}])

    # Explicitly set string_to32=True
    ak.to_parquet(a, tmp_path / "test_explicit_string_to32.parquet", string_to32=True)
    x = ak.from_parquet(tmp_path / "test_explicit_string_to32.parquet")

    # Convert to numpy (should use int32 offset kernels)
    result = np.asarray(x["name"])
    assert result.dtype.kind == "U"
    assert list(result) == ["Alice", "Bob", "Charlie"]


def test_parquet_string_roundtrip_string_to32_false(tmp_path):
    """Test parquet roundtrip with string_to32=False."""
    a = ak.Array([{"text": "hello"}, {"text": "world"}])

    # Set string_to32=False to use int64 offsets
    ak.to_parquet(a, tmp_path / "test_string_to32_false.parquet", string_to32=False)
    x = ak.from_parquet(tmp_path / "test_string_to32_false.parquet")

    # Convert to numpy (may use int64 offset kernels)
    result = np.asarray(x["text"])
    assert result.dtype.kind == "U"
    assert list(result) == ["hello", "world"]


def test_parquet_bytestring_roundtrip_with_bytestring_to32(tmp_path):
    """Test that bytestrings work correctly with bytestring_to32=True."""
    # Create array with bytestrings
    a = ak.Array([{"data": b"foo"}, {"data": b"bar"}])

    # Serialize with bytestring_to32=True (default)
    ak.to_parquet(a, tmp_path / "test_bytestring_to32.parquet", bytestring_to32=True)
    x = ak.from_parquet(tmp_path / "test_bytestring_to32.parquet")

    # Convert to numpy
    result = np.asarray(x["data"])
    assert result.dtype.kind == "S"
    assert list(result) == [b"foo", b"bar"]


def test_direct_string_array_with_int32_offsets(tmp_path):
    """Test direct conversion of string arrays with int32 offsets to NumPy."""
    # Create a string array
    strings = ak.Array(["hello", "world", "test"])

    # Serialize and deserialize through parquet to get int32 offsets
    ak.to_parquet(strings, tmp_path / "test_direct_strings.parquet", string_to32=True)
    x = ak.from_parquet(tmp_path / "test_direct_strings.parquet")

    # Check that offsets are int32
    layout = x.layout
    assert layout.offsets.dtype in (
        np.dtype("int32"),
        np.dtype("uint32"),
        np.dtype("int64"),
    )

    # Convert to numpy
    result = np.asarray(x)
    assert result.dtype.kind == "U"
    assert list(result) == ["hello", "world", "test"]


def test_empty_strings_with_int32_offsets(tmp_path):
    """Test that empty strings work correctly with int32 offsets."""
    a = ak.Array([{"msg": ""}, {"msg": "x"}, {"msg": ""}])

    ak.to_parquet(a, tmp_path / "test_empty_strings.parquet", string_to32=True)
    x = ak.from_parquet(tmp_path / "test_empty_strings.parquet")

    result = np.asarray(x["msg"])
    assert result.dtype.kind == "U"
    assert list(result) == ["", "x", ""]


def test_unicode_strings_with_int32_offsets(tmp_path):
    """Test that unicode strings work correctly with int32 offsets."""
    a = ak.Array([{"text": "Hello 👋"}, {"text": "World 🌍"}, {"text": "Test 🧪"}])

    ak.to_parquet(a, tmp_path / "test_unicode_strings.parquet", string_to32=True)
    x = ak.from_parquet(tmp_path / "test_unicode_strings.parquet")

    result = np.asarray(x["text"])
    assert result.dtype.kind == "U"
    assert list(result) == ["Hello 👋", "World 🌍", "Test 🧪"]



from __future__ import annotations

import pytest
from packaging.version import parse as parse_version

import awkward as ak

cudf = pytest.importorskip("cudf", exc_type=ImportError)
cupy = pytest.importorskip("cupy")


@pytest.mark.xfail(
    parse_version(cudf.__version__) >= parse_version("25.12.00"),
    reason="cudf internals changed since v25.12.00",
)
def test_jagged():
    arr = ak.Array([[[1, 2, 3], [], [3, 4]], []])
    out = ak.to_cudf(arr)
    assert isinstance(out, cudf.Series)
    assert out.to_arrow().tolist() == [[[1, 2, 3], [], [3, 4]], []]


@pytest.mark.xfail(
    parse_version(cudf.__version__) >= parse_version("25.12.00"),
    reason="cudf internals changed since v25.12.00",
)
def test_nested():
    arr = ak.Array(
        [{"a": 0, "b": 1.0, "c": {"d": 0}}, {"a": 1, "b": 0.0, "c": {"d": 1}}]
    )
    out = ak.to_cudf(arr)
    assert isinstance(out, cudf.Series)
    assert out.to_arrow().tolist() == [
        {"a": 0, "b": 1.0, "c": {"d": 0}},
        {"a": 1, "b": 0.0, "c": {"d": 1}},
    ]


@pytest.mark.xfail(
    parse_version(cudf.__version__) >= parse_version("25.12.00"),
    reason="cudf internals changed since v25.12.00",
)
def test_null():
    arr = ak.Array([12, None, 21, 12])
    # calls ByteMaskedArray._to_cudf not NumpyArray
    out = ak.to_cudf(arr)
    assert isinstance(out, cudf.Series)
    assert out.to_arrow().tolist() == [12, None, 21, 12]

    # True is valid, LSB order
    arr2 = ak.Array(arr.layout.to_BitMaskedArray(True, True))
    out = ak.to_cudf(arr2)
    assert isinstance(out, cudf.Series)
    assert out.to_arrow().tolist() == [12, None, 21, 12]

    # reversed LSB (should be rare, involves extra work!)
    arr3 = ak.Array(arr.layout.to_BitMaskedArray(True, False))
    out = ak.to_cudf(arr3)
    assert isinstance(out, cudf.Series)
    assert out.to_arrow().tolist() == [12, None, 21, 12]


@pytest.mark.xfail(
    parse_version(cudf.__version__) >= parse_version("25.12.00"),
    reason="cudf internals changed since v25.12.00",
)
def test_strings():
    arr = ak.Array(["hey", "hi", "hum"])
    out = ak.to_cudf(arr)
    assert out.to_arrow().tolist() == ["hey", "hi", "hum"]

    arr = ak.Array(["hey", "hi", None, "hum"])
    out = ak.to_cudf(arr)
    assert out.to_arrow().tolist() == ["hey", "hi", None, "hum"]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import os

import pytest

import awkward as ak

pytest.importorskip("pyarrow.parquet")


def test(tmp_path):
    filename = os.path.join(tmp_path, "whatever.parquet")

    px = ak.Array([[1.0, 2.0], [3.0], [], [4.0, 5.0, 6.0]])
    py = ak.Array([[0.5, 1.5], [2.5], [], [3.5, 4.5, 5.5]])
    pz = ak.Array([[0.2, 1.2], [2.2], [], [3.2, 4.2, 5.2]])
    e = ak.Array([[0.1, 1.1], [2.1], [], [3.1, 4.1, 5.1]])
    data = {"px": px, "py": py, "pz": pz, "e": e}  # not quite a record

    ak.to_parquet(data, filename)

    assert ak.from_parquet(filename).tolist() == ak.to_list(ak.to_layout(data))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest

import awkward as ak

jax = pytest.importorskip("jax")


def test():
    ak.jax.register_and_check()

    array = ak.Array([[1.0, 2.0, 3.0], [], [4.0, 5.0]], backend="jax")

    val_mean, grad_mean = jax.value_and_grad(ak.mean, argnums=0)(array)
    _, _grad_sum = jax.value_and_grad(ak.sum, argnums=0)(array)

    assert val_mean == 3
    assert ak.all(
        grad_mean == ak.Array([[0.2, 0.2, 0.2], [], [0.2, 0.2]], backend="jax")
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np

from awkward._nplikes.typetracer import TypeTracer, TypeTracerArray

nplike = TypeTracer.instance()


def test_all():
    buffer = TypeTracerArray._new(np.dtype("float32"), (4,))
    result = nplike.all(buffer, axis=None)
    assert isinstance(result, TypeTracerArray)
    assert result.dtype == np.dtype("bool")
    assert result.shape == ()

    buffer = TypeTracerArray._new(np.dtype("float32"), (4, 3))
    result = nplike.all(buffer, axis=None)
    assert isinstance(result, TypeTracerArray)
    assert result.dtype == np.dtype("bool")
    assert result.shape == ()


def test_any():
    buffer = TypeTracerArray._new(np.dtype("float32"), (4,))
    result = nplike.any(buffer, axis=None)
    assert isinstance(result, TypeTracerArray)
    assert result.dtype == np.dtype("bool")
    assert result.shape == ()

    buffer = TypeTracerArray._new(np.dtype("float32"), (4, 3))
    result = nplike.any(buffer, axis=None)
    assert isinstance(result, TypeTracerArray)
    assert result.dtype == np.dtype("bool")
    assert result.shape == ()



from __future__ import annotations

import cupy as cp
import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


def test_0315_integerindex_null_more():
    f = ak.highlevel.Array([[0, None, 2], None, [3, 4], []], backend="cuda").layout
    g1 = ak.highlevel.Array([[1, 2, None], None, [], [None]], backend="cuda").layout
    g2 = ak.highlevel.Array([[], None, None, []], backend="cuda").layout
    g3 = ak.highlevel.Array([[], [], [], []], backend="cuda").layout

    assert to_list(f[g1]) == [[None, 2, None], None, [], [None]]
    assert to_list(f[g2]) == [[], None, None, []]
    assert to_list(f[g3]) == [[], None, [], []]
    assert f.to_typetracer()[g1].form == f[g1].form
    assert f.to_typetracer()[g2].form == f[g2].form
    assert f.to_typetracer()[g3].form == f[g3].form

    a = ak.highlevel.Array([[0, 1, 2, None], None], backend="cuda").layout
    b = ak.highlevel.Array([[2, 1, None, 3], None], backend="cuda").layout

    assert to_list(a[b]) == [[2, 1, None, None], None]
    assert a.to_typetracer()[b].form == a[b].form

    b = ak.highlevel.Array([[2, 1, None, 3], []], backend="cuda").layout

    assert to_list(a[b]) == [[2, 1, None, None], None]
    assert a.to_typetracer()[b].form == a[b].form

    b = ak.highlevel.Array([[2, 1, None, 3], [0, 1]], backend="cuda").layout
    assert to_list(a[b]) == [[2, 1, None, None], None]
    assert a.to_typetracer()[b].form == a[b].form


def test_0315_integerindex_null_more_2():
    a = ak.highlevel.Array(
        [[[0, 1, 2, None], None], [[3, 4], [5]], None, [[6]]], backend="cuda"
    ).layout
    b = ak.highlevel.Array(
        [[[2, 1, None, 3], [0, 1]], [[0], None], None, [None]], backend="cuda"
    ).layout
    c = ak.highlevel.Array(
        [
            [[False, True, None, False], [False, True]],
            [[True, False], None],
            None,
            [None],
        ],
        backend="cuda",
    ).layout

    assert to_list(a[b]) == [
        [[2, 1, None, None], None],
        [[3], None],
        None,
        [None],
    ]
    assert to_list(a[c]) == [[[1, None], None], [[4], None], None, [None]]
    assert a.to_typetracer()[c].form == a[c].form


def test_1405_slicing_untested_cases_list_option_list():
    """Check that non-offset list(option(list indexes correctly"""
    content = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 2], dtype=np.int64)),
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([2, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([2, 2, 2], dtype=np.int64)),
        ),
    )

    index = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 2], dtype=np.int64)),
        ak.contents.IndexedOptionArray(
            ak.index.Index64(np.array([0, 1], dtype=np.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64(np.array([0, 0, 1], dtype=np.int64)),
                ak.contents.NumpyArray(np.array([0], dtype=np.int64)),
            ),
        ),
    )

    cuda_content = ak.to_backend(content, "cuda")
    cuda_index = ak.to_backend(index, "cuda")

    assert cuda_content[cuda_index].to_list() == [[[], [2]]]


def test_1405_slicing_untested_cases_list_option_list_offset():
    """Check that offset list(option(list indexes correctly"""
    content = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([1, 3], dtype=np.int64)),
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(np.array([2, 2, 2], dtype=np.int64)),
        ),
    )

    index = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 2], dtype=np.int64)),
        ak.contents.IndexedOptionArray(
            ak.index.Index64(np.array([0, 1], dtype=np.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64(np.array([0, 0, 1], dtype=np.int64)),
                ak.contents.NumpyArray(np.array([0], dtype=np.int64)),
            ),
        ),
    )

    cuda_content = ak.to_backend(content, "cuda")
    cuda_index = ak.to_backend(index, "cuda")

    assert cuda_content[cuda_index].to_list() == [[[], [2]]]


def test_1502_getitem_jagged_issue1406():
    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([1, 3], dtype=np.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64(np.array([0, 2, 2, 3], dtype=np.int64)),
                ak.contents.NumpyArray(np.array([0, 1, 2], dtype=np.int64)),
            ),
        ),
        check_valid=True,
    )

    index = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2], dtype=np.int64)),
            ak.contents.IndexedOptionArray(
                ak.index.Index64(np.array([0, 1], dtype=np.int64)),
                ak.contents.ListOffsetArray(
                    ak.index.Index64(np.array([0, 0, 1], dtype=np.int64)),
                    ak.contents.NumpyArray(np.array([0], dtype=np.int64)),
                ),
            ),
        ),
        check_valid=True,
    )

    cuda_array = ak.to_backend(array, "cuda")
    cuda_index = ak.to_backend(index, "cuda")

    assert to_list(cuda_array[cuda_index]) == [[[], [2]]]


def test_1502_getitem_jagged_issue1406_success_start_offset0():
    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2], dtype=np.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64(np.array([2, 2, 3], dtype=np.int64)),
                ak.contents.NumpyArray(np.array([0, 1, 2], dtype=np.int64)),
            ),
        ),
        check_valid=True,
    )

    index = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2], dtype=np.int64)),
            ak.contents.IndexedOptionArray(
                ak.index.Index64(np.array([0, 1], dtype=np.int64)),
                ak.contents.ListOffsetArray(
                    ak.index.Index64(np.array([0, 0, 1], dtype=np.int64)),
                    ak.contents.NumpyArray(np.array([0], dtype=np.int64)),
                ),
            ),
        ),
        check_valid=True,
    )

    cuda_array = ak.to_backend(array, "cuda")
    cuda_index = ak.to_backend(index, "cuda")

    assert to_list(cuda_array[cuda_index]) == [[[], [2]]]


def test_1502_getitem_jagged_issue1406_success_remove_option_type():
    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([1, 3], dtype=np.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64(np.array([0, 2, 2, 3], dtype=np.int64)),
                ak.contents.NumpyArray(np.array([0, 1, 2], dtype=np.int64)),
            ),
        ),
        check_valid=True,
    )

    index = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2], dtype=np.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64(np.array([0, 0, 1], dtype=np.int64)),
                ak.contents.NumpyArray(np.array([0], dtype=np.int64)),
            ),
        ),
        check_valid=True,
    )

    cuda_array = ak.to_backend(array, "cuda")
    cuda_index = ak.to_backend(index, "cuda")

    assert to_list(cuda_array[cuda_index]) == [[[], [2]]]


def test_1502_getitem_jagged_issue1406_success_nonempty_list():
    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([1, 3], dtype=np.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64(np.array([0, 1, 2, 3], dtype=np.int64)),
                ak.contents.NumpyArray(np.array([0, 1, 2], dtype=np.int64)),
            ),
        ),
        check_valid=True,
    )

    index = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2], dtype=np.int64)),
            ak.contents.IndexedOptionArray(
                ak.index.Index64(np.array([0, 1], dtype=np.int64)),
                ak.contents.ListOffsetArray(
                    ak.index.Index64(np.array([0, 1, 2], dtype=np.int64)),
                    ak.contents.NumpyArray(np.array([0, 0], dtype=np.int64)),
                ),
            ),
        ),
        check_valid=True,
    )

    cuda_array = ak.to_backend(array, "cuda")
    cuda_index = ak.to_backend(index, "cuda")

    assert to_list(cuda_array[cuda_index]) == [[[1], [2]]]


def test_1904_drop_none_ListArray_and_axis_None():
    index = ak.index.Index64(np.asarray([0, -1, 1, -1, 4, -1, 5]))
    content = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(
                np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])
            )
        ],
        ["nest"],
    )
    indexoptionarray = ak.contents.IndexedOptionArray(index, content)
    a = ak.contents.listarray.ListArray(
        ak.index.Index(np.array([4, 100, 1], np.int64)),
        ak.index.Index(np.array([7, 100, 3, 200], np.int64)),
        indexoptionarray,
    )

    cuda_a = ak.to_backend(a, "cuda")

    assert (
        to_list(ak.drop_none(cuda_a))
        == to_list(cuda_a[~ak.is_none(cuda_a, axis=1)])
        == [
            [{"nest": 4.4}, {"nest": 5.5}],
            [],
            [{"nest": 1.1}],
        ]
    )


def test_1904_drop_none_ListOffsetArray_IndexedOptionArray_NumpyArray_outoforder():
    index = ak.index.Index64(np.asarray([0, -1, 1, 5, 4, 2, 5]))
    content = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
    )
    indexoptionarray = ak.contents.IndexedOptionArray(index, content)
    offsets = ak.index.Index64(np.asarray([0, 4, 5, 6]))
    listoffset = ak.contents.ListOffsetArray(offsets, indexoptionarray)

    cuda_listoffset = ak.to_backend(listoffset, "cuda")

    assert to_list(cuda_listoffset) == [[0.0, None, 1.1, 5.5], [4.4], [2.2]]
    assert (
        to_list(ak.drop_none(cuda_listoffset, axis=1))
        == to_list(cuda_listoffset[~ak.is_none(cuda_listoffset, axis=1)])
        == [[0.0, 1.1, 5.5], [4.4], [2.2]]
    )
    assert to_list(ak.drop_none(cuda_listoffset)) == [[0.0, 1.1, 5.5], [4.4], [2.2]]


def test_1904_drop_none_from_iter():
    a = ak.Array([[1], [2, None]], backend="cuda")
    assert to_list(ak.drop_none(a)) == [[1], [2]]

    a = ak.Array([[2, None]], backend="cuda")
    assert to_list(ak.drop_none(a)) == [[2]]

    a = ak.Array([[[None]]], backend="cuda")
    assert to_list(ak.drop_none(a)) == [[[]]]

    a = ak.Array([1, 2, None], backend="cuda")
    assert to_list(ak.drop_none(a, axis=0))

    a = ak.Array([[[1, None]], [[3, 4]], [[5, 6]], [[7.8]]], backend="cuda")
    assert (
        to_list(ak.drop_none(a, axis=2))
        == to_list(a[~ak.is_none(a, axis=2)])
        == [[[1.0]], [[3.0, 4.0]], [[5.0, 6.0]], [[7.8]]]
    )

    a = ak.Array([[[0]], [[None]], [[1], None], [[2, None]]], backend="cuda")
    assert (
        to_list(ak.drop_none(a, axis=1))
        == to_list(a[~ak.is_none(a, axis=1)])
        == [[[0]], [[None]], [[1]], [[2, None]]]
    )

    a = ak.Array(
        [[[0]], [None, 34], [[1], None, 31], [[2, [[None]]]], [[[None]]]],
        backend="cuda",
    )
    assert (
        to_list(ak.drop_none(a, axis=0))
        == to_list(a[~ak.is_none(a, axis=0)])
        == [[[0]], [None, 34], [[1], None, 31], [[2, [[None]]]], [[[None]]]]
    )

    a = ak.Array([[[1, None]], [[3, None]], [[5, 6]], [[7.8]]], backend="cuda")
    assert (
        to_list(ak.drop_none(a, axis=2))
        == to_list(a[~ak.is_none(a, axis=2)])
        == [[[1.0]], [[3.0]], [[5.0, 6.0]], [[7.8]]]
    )

    a = ak.Array([[[1, None]], [[None, 4]], [[5, 6]], [[7.8]]], backend="cuda")
    assert (
        to_list(ak.drop_none(a, axis=2))
        == to_list(a[~ak.is_none(a, axis=2)])
        == [[[1.0]], [[4.0]], [[5.0, 6.0]], [[7.8]]]
    )

    a = ak.Array([[[1, None]], [[None, None]], [[5, 6]], [[7.8]]], backend="cuda")
    assert (
        to_list(ak.drop_none(a, axis=2))
        == to_list(a[~ak.is_none(a, axis=2)])
        == [[[1.0]], [[]], [[5.0, 6.0]], [[7.8]]]
    )

    a = ak.Array([[[1, None]], [[None, None]], [[None, 6]], [[7.8]]], backend="cuda")
    assert (
        to_list(ak.drop_none(a, axis=2))
        == to_list(a[~ak.is_none(a, axis=2)])
        == [[[1.0]], [[]], [[6.0]], [[7.8]]]
    )

    a = ak.Array(
        [[{"x": [1], "y": [[2]]}], [{"x": [None], "y": [[None]]}], None], backend="cuda"
    )
    assert to_list(a) == [
        [{"x": [1], "y": [[2]]}],
        [{"x": [None], "y": [[None]]}],
        None,
    ]
    assert to_list(ak.drop_none(a)) == [
        [{"x": [1], "y": [[2]]}],
        [{"x": [], "y": [[]]}],
    ]
    assert (
        to_list(ak.drop_none(a, axis=0))
        == to_list(a[~ak.is_none(a, axis=0)])
        == [[{"x": [1], "y": [[2]]}], [{"x": [None], "y": [[None]]}]]
    )
    assert to_list(ak.drop_none(a, axis=1)) == [
        [{"x": [1], "y": [[2]]}],
        [{"x": [], "y": [[None]]}],
        None,
    ]


def test_1904_drop_none_List_ByteMaskedArray_NumpyArray():
    a = ak.contents.listarray.ListArray(
        ak.index.Index(np.array([1, 3], np.int64)),
        ak.index.Index(np.array([3, 4], np.int64)),
        ak.contents.bytemaskedarray.ByteMaskedArray(
            ak.index.Index(np.array([1, 0, 1, 0, 1], np.int8)),
            ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
            valid_when=True,
        ),
    )

    cuda_a = ak.to_backend(a, "cuda")

    assert to_list(cuda_a) == [[None, 3.3], [None]]
    assert to_list(ak.drop_none(cuda_a)) == [[3.3], []]
    assert to_list(ak.drop_none(cuda_a, axis=1)) == to_list(
        cuda_a[~ak.is_none(cuda_a, axis=1)]
    )


def test_1904_drop_none_RegularArray_RecordArray_NumpyArray():
    index = ak.index.Index64(np.asarray([0, -1, 1, 2, 3, 4, -1, 6, 7, 8, -1, 10]))
    content = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.10])
    )
    indexoptionarray = ak.contents.IndexedOptionArray(index, content)
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.contents.listoffsetarray.ListOffsetArray(
        offsets, indexoptionarray
    )
    regulararray = ak.contents.regulararray.RegularArray(listoffsetarray, 2)

    cuda_regulararray = ak.to_backend(regulararray, "cuda")

    assert to_list(cuda_regulararray) == [
        [[0.0, None, 1.1], []],
        [[2.2, 3.3], [4.4]],
        [[None, 6.6, 7.7, 8.8], []],
    ]
    assert to_list(ak.drop_none(cuda_regulararray, axis=2)) == to_list(
        cuda_regulararray[~ak.is_none(cuda_regulararray, axis=2)]
    )


def test_1904_drop_none_RecordArray():
    a = ak.Array(
        [
            [{"x": [1], "y": [[2]]}],
            None,
            [None],
            [{"x": None, "y": None}],
            [{"x": [None], "y": [None]}],
            [{"x": [11], "y": [[None]]}],
        ],
        backend="cuda",
    )
    assert to_list(ak.drop_none(a, axis=1)) == to_list(a[~ak.is_none(a, axis=1)])
    assert to_list(ak.drop_none(a, axis=2)) == [
        [{"x": [1], "y": [[2]]}],
        None,
        [None],
        [{"x": None, "y": None}],
        [{"x": [], "y": []}],
        [{"x": [11], "y": [[None]]}],
    ]


def test_simple_slice():
    arrg = ak.Array([[1, 2, 3], [0], [4, 5]], backend="cuda")
    out = arrg[:, 0]
    assert out.tolist() == [1, 0, 4]


def test_2246_slice_not_packed():
    index = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64([0, 3, 5]),
            ak.contents.NumpyArray(
                np.array([True, False, False, True, True, False, False], dtype=np.bool_)
            ),
        )
    )
    array = ak.Array([[0, 1, 2], [3, 4]])

    cuda_index = ak.to_backend(index, "cuda")
    cuda_array = ak.to_backend(array, "cuda")

    result = cuda_array[cuda_index]
    assert result.tolist() == [[0], [3, 4]]


def test_0127_tomask_operation_ByteMaskedArray_jaggedslice0():
    array = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    index = ak.index.Index64(np.array([0, 1, 2, 3], dtype=np.int64))
    indexedarray = ak.contents.IndexedOptionArray(index, array)

    cuda_indexedarray = ak.to_backend(indexedarray, "cuda")

    assert to_list(cuda_indexedarray) == [
        [0.0, 1.1, 2.2],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert to_list(
        cuda_indexedarray[
            ak.highlevel.Array([[0, -1], [0], [], [1, 1]], backend="cuda")
        ]
    ) == [
        [0.0, 2.2],
        [3.3],
        [],
        [7.7, 7.7],
    ]

    mask = ak.index.Index8(np.array([0, 0, 0, 0], dtype=np.int8))
    maskedarray = ak.contents.ByteMaskedArray(mask, array, valid_when=False)

    cuda_maskedarray = ak.to_backend(maskedarray, "cuda")

    assert to_list(cuda_maskedarray) == [
        [0.0, 1.1, 2.2],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert to_list(
        cuda_maskedarray[ak.highlevel.Array([[0, -1], [0], [], [1, 1]], backend="cuda")]
    ) == [
        [0.0, 2.2],
        [3.3],
        [],
        [7.7, 7.7],
    ]


def test_0127_tomask_operation_ByteMaskedArray_jaggedslice1():
    model = ak.highlevel.Array(
        [
            [0.0, 1.1, None, 2.2],
            [],
            [3.3, None, 4.4],
            [5.5],
            [6.6, 7.7, None, 8.8, 9.9],
        ],
        backend="cuda",
    )
    assert to_list(
        model[
            ak.highlevel.Array(
                [[3, 2, 1, 1, 0], [], [1], [0, 0], [1, 2]], backend="cuda"
            )
        ]
    ) == [
        [2.2, None, 1.1, 1.1, 0.0],
        [],
        [None],
        [5.5, 5.5],
        [7.7, None],
    ]

    content = ak.contents.NumpyArray(
        np.array([0.0, 1.1, 999, 2.2, 3.3, 123, 4.4, 5.5, 6.6, 7.7, 321, 8.8, 9.9])
    )
    mask = ak.index.Index8(
        np.array([0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0], dtype=np.int8)
    )
    maskedarray = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    offsets = ak.index.Index64(np.array([0, 4, 4, 7, 8, 13], dtype=np.int64))
    listarray = ak.highlevel.Array(
        ak.contents.ListOffsetArray(offsets, maskedarray), backend="cuda"
    )

    assert to_list(listarray) == to_list(model)
    assert to_list(
        listarray[
            ak.highlevel.Array(
                [[3, 2, 1, 1, 0], [], [1], [0, 0], [1, 2]], backend="cuda"
            )
        ]
    ) == [[2.2, None, 1.1, 1.1, 0.0], [], [None], [5.5, 5.5], [7.7, None]]


def test_0127_tomask_operation_ByteMaskedArray_jaggedslice2():
    model = ak.highlevel.Array(
        [
            [[0.0, 1.1, None, 2.2], [], [3.3, None, 4.4]],
            [],
            [[5.5]],
            [[6.6, 7.7, None, 8.8, 9.9]],
        ],
        backend="cuda",
    )
    assert to_list(
        model[
            ak.highlevel.Array(
                [[[3, 2, 1, 1, 0], [], [1]], [], [[0, 0]], [[1, 2]]], backend="cuda"
            )
        ]
    ) == [[[2.2, None, 1.1, 1.1, 0.0], [], [None]], [], [[5.5, 5.5]], [[7.7, None]]]

    content = ak.contents.NumpyArray(
        np.array([0.0, 1.1, 999, 2.2, 3.3, 123, 4.4, 5.5, 6.6, 7.7, 321, 8.8, 9.9])
    )
    mask = ak.index.Index8(
        np.array([0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0], dtype=np.int8)
    )
    maskedarray = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    offsets = ak.index.Index64(np.array([0, 4, 4, 7, 8, 13], dtype=np.int64))
    sublistarray = ak.contents.ListOffsetArray(offsets, maskedarray)
    offsets2 = ak.index.Index64(np.array([0, 3, 3, 4, 5], dtype=np.int64))
    listarray = ak.highlevel.Array(
        ak.contents.ListOffsetArray(offsets2, sublistarray), backend="cuda"
    )
    assert to_list(listarray) == to_list(model)
    assert to_list(
        listarray[
            ak.highlevel.Array(
                [[[3, 2, 1, 1, 0], [], [1]], [], [[0, 0]], [[1, 2]]], backend="cuda"
            )
        ]
    ) == [[[2.2, None, 1.1, 1.1, 0.0], [], [None]], [], [[5.5, 5.5]], [[7.7, None]]]


def test_0127_tomask_operation_ByteMaskedArray_jaggedslice3():
    model = ak.highlevel.Array(
        [
            [[[0.0, 1.1, None, 2.2], [], [3.3, None, 4.4]], []],
            [[[5.5]], [[6.6, 7.7, None, 8.8, 9.9]]],
        ],
        backend="cuda",
    )
    assert to_list(
        model[
            ak.highlevel.Array(
                [[[[3, 2, 1, 1, 0], [], [1]], []], [[[0, 0]], [[1, 2]]]], backend="cuda"
            )
        ]
    ) == [[[[2.2, None, 1.1, 1.1, 0.0], [], [None]], []], [[[5.5, 5.5]], [[7.7, None]]]]

    content = ak.contents.NumpyArray(
        np.array([0.0, 1.1, 999, 2.2, 3.3, 123, 4.4, 5.5, 6.6, 7.7, 321, 8.8, 9.9])
    )
    mask = ak.index.Index8(
        np.array([0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0], dtype=np.int8)
    )
    maskedarray = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    offsets = ak.index.Index64(np.array([0, 4, 4, 7, 8, 13], dtype=np.int64))
    subsublistarray = ak.contents.ListOffsetArray(offsets, maskedarray)
    offsets2 = ak.index.Index64(np.array([0, 3, 3, 4, 5], dtype=np.int64))
    sublistarray = ak.contents.ListOffsetArray(offsets2, subsublistarray)
    offsets3 = ak.index.Index64(np.array([0, 2, 4], dtype=np.int64))
    listarray = ak.highlevel.Array(
        ak.contents.ListOffsetArray(offsets3, sublistarray), backend="cuda"
    )
    assert to_list(listarray) == to_list(model)
    assert to_list(
        listarray[
            ak.highlevel.Array(
                [[[[3, 2, 1, 1, 0], [], [1]], []], [[[0, 0]], [[1, 2]]]], backend="cuda"
            )
        ]
    ) == [[[[2.2, None, 1.1, 1.1, 0.0], [], [None]], []], [[[5.5, 5.5]], [[7.7, None]]]]


def test_0127_tomask_operation():
    array = ak.highlevel.Array(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], backend="cuda"
    )
    mask1 = ak.highlevel.Array([True, True, False, False, True], backend="cuda")
    assert to_list(array[mask1]) == [[0.0, 1.1, 2.2], [], [6.6, 7.7, 8.8, 9.9]]
    assert to_list(ak.operations.mask(array, mask1)) == [
        [0.0, 1.1, 2.2],
        [],
        None,
        None,
        [6.6, 7.7, 8.8, 9.9],
    ]

    mask2 = ak.highlevel.Array(
        [[False, True, False], [], [True, True], [False], [True, False, False, True]],
        backend="cuda",
    )
    assert to_list(array[mask2]) == [[1.1], [], [3.3, 4.4], [], [6.6, 9.9]]
    assert to_list(ak.operations.mask(array, mask2)) == [
        [None, 1.1, None],
        [],
        [3.3, 4.4],
        [None],
        [6.6, None, None, 9.9],
    ]


def test_simple_slice_cpu():
    arr = ak.Array([[1, 2, 3], [0], [4, 5]])
    out = arr[:, 0]
    expected = [1, 0, 4]
    result = out.tolist()
    cp.testing.assert_array_list_equal(
        result,
        expected,
        err_msg=f"Slice of [[1, 2, 3], [0], [4, 5]] should be {expected}, but got {result}",
    )


def test_simple_slice_gpu():
    arr = ak.Array([[1, 2, 3], [0], [4, 5]], backend="cuda")
    out = arr[:, 0]
    expected = [1, 0, 4]
    result = out.tolist()
    cp.testing.assert_array_list_equal(
        result,
        expected,
        err_msg=f"Slice of [[1, 2, 3], [0], [4, 5]] should be {expected}, but got {result}",
    )


def test_simple_slice_cpu1():
    arr = ak.Array([[1, 2, 3], [0], [4, 5]])
    out = arr[:, 1:]
    expected = [[2, 3], [], [5]]
    result = out.tolist()
    cp.testing.assert_array_list_equal(
        result,
        expected,
        err_msg=f"Slice of [[1, 2, 3], [0], [4, 5]] should be {expected}, but got {result}",
    )


def test_simple_slice_gpu1():
    arr = ak.Array([[1, 2, 3], [0], [4, 5]], backend="cuda")
    out = arr[:, 1:]
    expected = [[2, 3], [], [5]]
    result = out.tolist()
    cp.testing.assert_array_list_equal(
        result,
        expected,
        err_msg=f"Slice of [[1, 2, 3], [0], [4, 5]] should be {expected}, but got {result}",
    )


def test_simple_slice_cpu2():
    arr = ak.Array([[1, 2, 3], [0], [4, 5]])
    out = arr[:, :1]
    expected = [[1], [0], [4]]
    result = out.tolist()
    cp.testing.assert_array_list_equal(
        result,
        expected,
        err_msg=f"Slice of [[1, 2, 3], [0], [4, 5]] should be {expected}, but got {result}",
    )


def test_simple_slice_gpu2():
    arr = ak.Array([[1, 2, 3], [0], [4, 5]], backend="cuda")
    out = arr[:, :1]
    expected = [[1], [0], [4]]
    result = out.tolist()
    cp.testing.assert_array_list_equal(
        result,
        expected,
        err_msg=f"Slice of [[1, 2, 3], [0], [4, 5]] should be {expected}, but got {result}",
    )


def test_simple_slice_cpu3():
    arr = ak.Array([[1, 2, 3], [0], [4, 5]])
    out = arr[:, 1::2]
    expected = [[2], [], [5]]
    result = out.tolist()
    cp.testing.assert_array_list_equal(
        result,
        expected,
        err_msg=f"Slice of [[1, 2, 3], [0], [4, 5]] should be {expected}, but got {result}",
    )


def test_simple_slice_gpu3():
    arr = ak.Array([[1, 2, 3], [0], [4, 5]], backend="cuda")
    out = arr[:, 1::2]
    expected = [[2], [], [5]]
    result = out.tolist()
    cp.testing.assert_array_list_equal(
        result,
        expected,
        err_msg=f"Slice of [[1, 2, 3], [0], [4, 5]] should be {expected}, but got {result}",
    )


def test_simple_slice_cpu4():
    arr = ak.Array([[1, 2, 3], [0], [4, 5]])
    out = arr[:, ::-1]
    expected = [[3, 2, 1], [0], [5, 4]]
    result = out.tolist()
    cp.testing.assert_array_list_equal(
        result,
        expected,
        err_msg=f"Slice of [[1, 2, 3], [0], [4, 5]] should be {expected}, but got {result}",
    )


def test_simple_slice_gpu4():
    arr = ak.Array([[1, 2, 3], [0], [4, 5]], backend="cuda")
    out = arr[:, ::-1]
    expected = [[3, 2, 1], [0], [5, 4]]
    result = out.tolist()
    cp.testing.assert_array_list_equal(
        result,
        expected,
        err_msg=f"Slice of [[1, 2, 3], [0], [4, 5]] should be {expected}, but got {result}",
    )



from __future__ import annotations

import numpy as np

import awkward as ak


def test():
    layout = ak.to_layout([[[1, 2, 3], [4]], None]).to_BitMaskedArray(
        valid_when=True, lsb_order=True
    )
    result = ak.flatten(layout, axis=1, highlevel=False)
    assert result.is_equal_to(
        ak.contents.ListOffsetArray(
            ak.index.Index64([0, 3, 4]),
            ak.contents.NumpyArray(np.array([1, 2, 3, 4], dtype=np.int64)),
        )
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import ast

import numpy as np  # noqa: F401

import awkward as ak

PREFIX = "foo"


def test_NumpyForm():
    form = ak.forms.NumpyForm("float64")
    assert form.form_key is None

    form_with_unique_key = ak.forms.form_with_unique_keys(form, (PREFIX,))
    assert ast.literal_eval(form_with_unique_key.form_key) == (PREFIX,)


def test_BitMaskedForm():
    form = ak.forms.BitMaskedForm(
        "i8",
        ak.forms.NumpyForm("float64"),
        True,
        False,
    )
    assert form.form_key is None

    form_with_unique_key = ak.forms.form_with_unique_keys(form, (PREFIX,))
    assert ast.literal_eval(form_with_unique_key.form_key) == (PREFIX,)
    assert ast.literal_eval(form_with_unique_key.content.form_key) == (PREFIX, None)


def test_EmptyForm():
    form = ak.forms.EmptyForm()
    assert form.form_key is None

    form_with_unique_key = ak.forms.form_with_unique_keys(form, (PREFIX,))
    assert ast.literal_eval(form_with_unique_key.form_key) == (PREFIX,)


def test_IndexedForm():
    form = ak.forms.IndexedForm(
        "i64",
        ak.forms.NumpyForm("float64"),
    )
    assert form.form_key is None

    form_with_unique_key = ak.forms.form_with_unique_keys(form, (PREFIX,))
    assert ast.literal_eval(form_with_unique_key.form_key) == (PREFIX,)
    assert ast.literal_eval(form_with_unique_key.content.form_key) == (PREFIX, None)


def test_IndexedOptionForm():
    form = ak.forms.IndexedOptionForm(
        "i64",
        ak.forms.NumpyForm("float64"),
    )
    assert form.form_key is None

    form_with_unique_key = ak.forms.form_with_unique_keys(form, (PREFIX,))
    assert ast.literal_eval(form_with_unique_key.form_key) == (PREFIX,)
    assert ast.literal_eval(form_with_unique_key.content.form_key) == (PREFIX, None)


def test_ListForm():
    form = ak.forms.ListForm(
        "i64",
        "i64",
        ak.forms.NumpyForm("float64"),
    )
    assert form.form_key is None

    form_with_unique_key = ak.forms.form_with_unique_keys(form, (PREFIX,))
    assert ast.literal_eval(form_with_unique_key.form_key) == (PREFIX,)
    assert ast.literal_eval(form_with_unique_key.content.form_key) == (PREFIX, None)


def test_ListOffsetForm():
    form = ak.forms.ListOffsetForm(
        "i64",
        ak.forms.NumpyForm("float64"),
    )
    assert form.form_key is None

    form_with_unique_key = ak.forms.form_with_unique_keys(form, (PREFIX,))
    assert ast.literal_eval(form_with_unique_key.form_key) == (PREFIX,)
    assert ast.literal_eval(form_with_unique_key.content.form_key) == (PREFIX, None)


def test_RecordForm():
    form = ak.forms.RecordForm(
        [ak.forms.NumpyForm("float64"), ak.forms.NumpyForm("bool")],
        ["one", "two"],
    )
    assert form.form_key is None

    form_with_unique_key = ak.forms.form_with_unique_keys(form, (PREFIX,))
    assert ast.literal_eval(form_with_unique_key.content("one").form_key) == (
        PREFIX,
        "one",
    )
    assert ast.literal_eval(form_with_unique_key.content("two").form_key) == (
        PREFIX,
        "two",
    )


def test_RegularForm():
    form = ak.forms.RegularForm(
        ak.forms.NumpyForm("float64"),
        10,
    )
    assert form.form_key is None

    form_with_unique_key = ak.forms.form_with_unique_keys(form, (PREFIX,))
    assert ast.literal_eval(form_with_unique_key.form_key) == (PREFIX,)
    assert ast.literal_eval(form_with_unique_key.content.form_key) == (PREFIX, None)


def test_UnionForm():
    form = ak.forms.UnionForm(
        "i8",
        "i64",
        [ak.forms.NumpyForm("float64"), ak.forms.NumpyForm("bool")],
    )
    assert form.form_key is None

    form_with_unique_key = ak.forms.form_with_unique_keys(form, (PREFIX,))
    assert ast.literal_eval(form_with_unique_key.contents[0].form_key) == (PREFIX, 0)
    assert ast.literal_eval(form_with_unique_key.contents[1].form_key) == (PREFIX, 1)


def test_UnmaskedForm():
    form = ak.forms.UnmaskedForm(
        ak.forms.NumpyForm("float64"),
    )
    assert form.form_key is None

    form_with_unique_key = ak.forms.form_with_unique_keys(form, (PREFIX,))
    assert ast.literal_eval(form_with_unique_key.form_key) == (PREFIX,)
    assert ast.literal_eval(form_with_unique_key.content.form_key) == (PREFIX, None)
